import unittest
from idlib.cache import cache_hash, argspector


class TestCacheHash(unittest.TestCase):
    def test_argspector(self):
        class c:
            def f(self, value): pass
            def g(self, value=None): pass
            def h(self, a, *b, c=None, **d): pass

        def f(value): pass
        def g(value=None): pass
        def h(a, *b, c=None, **d): pass

        ic = c()

        spectors = [
            [argspector(c.f), [((ic, 'f'), {}),]],
            [argspector(c.g), [((ic, 'g'), {}),
                               ((ic,), {'value':None}),]],
            [argspector(c.h), [((ic, 'h'), {}),
                               ((ic, 'a', 'b', 'oops'), dict(c='c', d='d', e='e')),]],
            [argspector(f), [(('f',), {}),]],
            [argspector(g), [(('g',), {}),
                             ([], {'value':None}),]],
            [argspector(h), [(('h',), {}),
                             (('a', 'b', 'oops'), dict(c='c', d='d', e='e')),]],
            ]


        for spector, arg_sets in spectors:
            for args, kwargs in arg_sets:
                pairs = list(spector(*args, **kwargs))
                cache_hash(pairs)



"""While the name of the class is Stream, the reality is much closer
to "a random block of data/code that may or may not be executable" in
the current runtime. The implication is mostly that it is possible to
serialize, save-lisp-and-die, etc. something and, given the right
receiving environment, recover it back to some Turing agentous state
if you so choose. Thus, in Python, Stream is just a regular old object
with a few restrictions on its structre.

Another reason why Stream is used as the name is because most of the
objects that are dereferenced to by persistent identifiers are readonly,
are only writeable by a select few people, or therwise change with a time
constant that is significantly longer than life cycle of code that consumes
them.
"""
# NOTE: at the moment, the implementaiton here is not fully homogenous.
# It is possible for methods to produce something other than another stream.
# pyontutils.core.OntRes is what we are shooting for, where an identifier
# identifies multiple streams collectively, they are most distingiushed by
# the particular type of substream, (e.g. metadata-free, data-homogenous)

from types import MappingProxyType
import idlib
from idlib import exceptions as exc
from idlib.utils import cache_result, StringProgenitor, log


# TODO it seems like there is a little dance going on between identifiers and local names
# and between identifiers and the actionable form
# local name < local conventions > canonical globally unique identifier < ??? > actionable form
# of course that global uniquesness is patently false, it is always global within some
# set of axioms, i.e. set of string translation rules

# Practically this suggests that it might be possible to set the 'action convetions'
# for identifiers in a way similar to the local conventions, maybe a set of sane
# defaults plus a 'get actionable form' or something?


class Stream:

    _id_class = None

    _progenitors = MappingProxyType({})  # this is a dict, but immutable to prevent accidents

    def __new__(cls, *args, **kwargs):
        return super().__new__(cls)

    _renew = __new__

    def __new__(cls, *args, **kwargs):
        Stream._setup(*args, **kwargs)
        Stream.__new__ = Stream._renew
        return super().__new__(cls)

    @staticmethod
    def _setup(*args, **kwargs):
        import requests  # resolver dejoure
        Stream._requests = requests

        from idlib.core import resolution_chain_responses
        Stream._resolution_chain_responses = staticmethod(resolution_chain_responses)

    @classmethod
    def fromJson(cls, blob):
        # TODO validation ...
        identifier = blob['id']
        if isinstance(identifier, cls):
            return identifier
        else:
            i = cls(identifier)
            i._progenitors = {}  # FIXME klobbered in self.data
            # TODO what is our term for the atomic part of a
            # stream into which no light may pass? it is different
            # than the homogenous data, it is more along the lines
            # of "this is some other data structure"
            i._progenitors['blob'] = blob
            return i

    @classmethod
    def fromIdInit(cls, *args, **kwargs):
        return cls(cls._id_class(*args, **kwargs))

    def __init__(self, identifier_or_id_as_string=None, *args, **kwargs):
        # FIXME or identifier, string, or another stream
        if isinstance(self, idlib.Identifier):
            # stream should always come last in the mro so it will hit object
            super().__init__() #identifier_or_id_as_string, *args, **kwargs)
            self._identifier = self._id_class(self)

        else:
            if isinstance(identifier_or_id_as_string, self._id_class):
                self._identifier = identifier_or_id_as_string
            else:
                if self._id_class == str and identifier_or_id_as_string is None:
                    # have to protect string identifiers because they can't type
                    # check their own inputs
                    raise TypeError('identifier_or_id_as_string may not be None')

                if isinstance(identifier_or_id_as_string, Stream):  # FIXME not quite right
                    self._identifier = identifier_or_id_as_string.identifier
                else:
                    self._identifier = self._id_class(identifier_or_id_as_string)

    @property
    def label(self):
        raise NotImplementedError('impl in subclass')

    @property
    def label_safe(self):
        try:
            return self.label
        except exc.RemoteError:
            return self.asStr()

    def asType(self, _class):
        # FIXME not quite ... but close
        if hasattr(self, '_id_class'):
            if self._id_class == str:
                return _class(self.identifier)
            else:
                return self.identifier.asType(_class)
        else:
            return _class(self.identifier)

    def asCell(self, sep='|'):
        """ As a cell in a table. """
        # TODO Distinct from asTabular in that asTabular is probably a row?
        if hasattr(self, '_id_class') and self._id_class:
            if self._id_class == str:
                local = self.identifier
            else:
                local = self.identifier.asLocal()
            try:
                if not self.label:
                    return local
            except (exc.ResolutionError, exc.RemoteError) as e:
                log.exception(e)
                return local
            except NotImplementedError as e:  # idlib.Identifier case
                log.error(f'self.label not implemented for {self.__class__}')
                log.exception(e)
                return local

            return self.label + sep + local

        else:
            # FIXME this needs some extension mechanism ...
            raise NotImplementedError('TODO as needed')

    def asDict(self, include_description=False):
        """ XXX this should NEVER allow an error to escape.
            Only return less information. """
        if hasattr(self, '_id_class') and self._id_class:
            out = {
                'type': 'identifier',
                'system': self.__class__.__name__,
                #'id': self.identifier,
                'id': self,  # XXX NOTE THE TRADEOFF HERE, this preserves stream access
                # 'alternate_identifiers': [],  # TODO
            }

            try:
                metadata = self.metadata()
            except exc.RemoteError as e:
                out['errors'] = [{'message': str(e)}]
                return out

            if metadata is not None:
                out['label'] = self.label
                if hasattr(self, 'category') and self.category:
                    out['category'] = self.category

                if hasattr(self, 'synonyms') and self.synonyms:
                    out['synonyms'] = self.synonyms

                if include_description and hasattr(self, 'description') and self.description:
                    out['description'] = self.description

            else:
                out['errors'] = [{'message': 'No metadata found.'}]

        # TODO other stream types e.g. pathlib.Path ... sigh CL MOP would be so useful here
        else:
            out = {'type': 'stream',
                   'system': self.__class__.__name__,  # FIXME key name ...
                   'id': self.identifier,}

        return out

    def __contains__(self, value):
        return value in self.identifier

    @property
    def identifier(self):
        """ Implicitly the unbound identifier that is to be dereferenced. """
        # FIXME interlex naming conventions call this a reference_name
        # in order to give it a bit more lexical distance from identity
        # which implies some hash function
        return self._identifier

    @identifier.setter
    def identifier(self, value):
        raise NotImplementedError('You probably meant to set self._identifier?')  # probably should never allow this ...

    def asStr(self):
        """ Return the idlib string representation of the identifier """
        # FIXME we probably need a safe version of this that
        # errors if the identifier/stream is a local identifier
        if hasattr(self, '_id_class') and self._id_class is str:
            return self.identifier
        #elif not hasattr(self, '_id_class'):  # will probably need this
        else:
            return self.identifier.asStr()

    def __eq__(self, other):
        return self.__class__ == other.__class__ and self.identifier == other.identifier

    def __hash__(self):
        """ The hash here identifies the NAME of the stream
            NOT the STATE of the stream. Two streams have the
            same hash if they have the same identifier. If you
            want to know if there are two equal streams that
            are different objects use self `is' other instead. """
        return hash((self.__class__, self.asStr()))

    @property
    def identifier_actionable(self):
        """ The 'resolvable' or actionable form of an identifier
            where actionable is defined by specification of the system
            NOT of the current context in which an identifier is being used """

        # in URI based systems this will just be the identifier
        # decoupling this important for allowing identifiers to
        # have canonical regex that are resolver independent

        # TODO identifier_actionable_resolver(self, resolver) or similar
        raise NotImplementedError('Implement in subclass')

    def checksum(self, cypher=None):  # FIXME default cypher value
        if not hasattr(self, '__checksum'):  # NOTE can set __checksum on the fly
            self.__checksum = self._checksum(cypher)

        return self.__checksum

    identity = checksum  # FIXME the naming here is mathematically incorrect

    def _checksum(self, cypher):
        raise NotImplementedError

    def progenitor(self, *args, level=1, type=None, _type=type):
        """ return the reproductible progenitor object/stream at level n
            also includes progenitors that are not reproducible but that
            retain metadata about the substream in question, e.g a requests
            response object has information about the superstream for a
            text stream

            generators and filelike objects should not be included in
            the progenitor list since they do not represent stateless
            progenitors (i.e. a generator can only be expressed once)

            note that sometimes this may include a tuple if the information
            needed has more than one part, no conventions for this have
            been decided yet

            level == 0 -> returns self to break"""

        if args:
            raise TypeError('progenitor accepts keywords arguments only!')

        # FIXME probably makes more sense to store progenitors in a dict
        # with a controlled set of types than just by accident of ordering
        # this would allow us to keep everything along the way
        # of course watch out for garbage collection issues ala lxml etree

        # XXX NOTE: self._progenitors = [] should be set EVERY time data is retrieved
        # it should NOT be set during __init__
        # if not hasattr(self, '_progenitors'):  # TODO

        # TODO instrumenting all the other objects with
        # a progenitor method is going to be a big sigh
        if type is not None:
            return self._progenitors[type]  # FIXME need enum or something

        if level is not None and level == 0:
            return self
        else:
            if level < 0:
                level += 1  # compensate for reversed

            # HACK to retain a stack of progenitors manually for now
            # FIXME blased insertion order :/ implemnation details
            return list(self._progenitors.values())[-level]

    # superstream = progenitor  # it would be level=1 but not as useful as we would like

    def dereference_chain(self):
        raise NotImplementedError

    def dereference(self, asType=None):
        """ Many identifier systems have native dereferincing semantics

            This particular dereferencing refers ONLY to the dereferencing
            of one identifier into another identifier stream, NOT the contents
            of that stream. This is the "default" dereferencing behavior of the
            _system_ NOT of the identifier. """

        # FIXME pretty sure that the behavior of the
        # dereferencing here is inconsistent and incorrect
        # many systems cannot dereference to their referent

        # as defined this is mostly used in the context
        # of dereferncing DNS/url resolution chains
        # which is useful but mostly an implementation detail

        # having explicit classes for substreams may help here
        # by making it possible to dereference the identifier,
        # the metadata stream, the data stream, etc. independenly
        raise NotImplementedError

    def headers(self):
        """ Data from the lower level stream from which this stream is
            derived/differentiated """
        # FIXME naming ...
        # sometimes this is related to ...
        #  transport
        #  prior in time
        #  unbound metadata, or metadata that will be unbound in the target
        #  metadata kept by the stream management layer (e.g. file system)
        #  stream type
        #  stream size
        #  operational metadata
        #  summary information

        # if you are sending a file then populate all the info
        # needed by the server to set up the stream (even if that seems a bit low level)
        raise NotImplementedError
        return self.superstream.metadata

    #@headers.setter
    #def headers(self, value):
        #self._headers = value
        #raise NotImplementedError('If you override self.headers in the child '
                                  #'you need to reimplement this too.')

    def data(self, mimetype_accept=None):
        """ The primary opaque datastream that will be differentiated
            at the next level. Identifiers for classes and pysical samples
            should return None or error on these. """
        raise NotImplementedError

    def metadata(self, mimetype_accept=None):
        """ stream metadata, hopefully as a header

            technically this should be called metadata_bound
            since it is part of the same stream, it might
            make sense to invert this to have a variety of
            datasources external to a stream that contain
            additional relevant data using that id
        """
        if not hasattr(self, '_metadata'):
            self._metadata = self._metadata_class(self.identifier)

        return self._metadata

    # TODO warn about deprecation for pyontutils

    @property
    def identifier_bound(self):
        raise NotImplementedError
        return self.metadata.identifier_bound

    @property
    def identifier_version(self):
        """ implicitly identifier_bound_version """
        raise NotImplementedError
        return self.metadata.identifier_version


### see docs for the trasition grammer


class MetadataFree(Stream):
    # NOT the primary referent of the identifier type in question
    pass


class DataFree(Stream):  # aka DataHeterogeneous
    # the primary referent of the identifier type in question
    pass


class MetadataBound(Stream):
    pass


class DataHomogenous(Stream):
    pass


class HelperNoData:  # FIXME should be
    def data(self):
        raise exc.IdentifierDoesNotDereferenceToDataError(self)

    @property
    def id_bound_data(self):
        raise exc.IdentifierDoesNotDereferenceToDataError(self)

    @property
    def _prgn_metadata(self):
        if hasattr(self, '_resp_metadata'):
            return self._resp_metadata
        elif hasattr(self, '_path_metadata'):
            return self._path_metadata
        else:
            raise AttributeError('no progenitor for metadata')

    def progenitor(self):
        metadata = self.metadata()
        # metadata.progenitor  # FIXME really prefer this ...
        return self.dereference_chain(), self._prgn_metadata  # FIXME assumptions about uri ...


class HelpTestStreams:
    stream = None
    ids = tuple()
    ids_bad = tuple()

    @staticmethod
    def setUpClass():
        if not hasattr(HelpTestStreams, '_requests'):
            import pytest
            HelpTestStreams._pytest = pytest
            import requests
            HelpTestStreams._requests = requests
            import hashlib
            HelpTestStreams._hashlib = hashlib
            import copy
            HelpTestStreams._copy = copy
            import pickle
            HelpTestStreams._pickle = pickle
            from joblib import Parallel, delayed
            HelpTestStreams._Parallel = Parallel
            HelpTestStreams._delayed = delayed

    @classmethod
    def _lol_joblib(cls, d):
        """ HEY KIDS WATCH THIS """
        cls._Parallel(n_jobs=2)(cls._delayed(lambda d: d)(d) for d in (d,))

    def test_stream_sections(self):
        # TODO run each on of the properties/methods in
        # a separate loop?
        cypher = self._hashlib.blake2b
        bads = []
        for i in self.ids:
            d = self.stream(i)
            d.identifier
            if d.identifier and hasattr(d.identifier, 'prefix'):
                print(d.identifier)
                print(d.identifier.prefix)
                print(d.identifier.suffix)
                if d.identifier.prefix is None or d.identifier.suffix is None:
                    bads.append(d.identifier.__dict__)

            try:
                d.identifier_bound_metadata
                d.identifier_bound_version_metadata

                d.checksum(cypher)  # data or metadata? indication that there should be multiple probably
                d.dereference()  # XXX induces infinite recursion
                d.headers()
                d.metadata()

                if not isinstance(d, HelperNoData):
                    d.data()

                d.progenitor()  # or is it d.data.progenitor?

                # test pickling
                hrm = self._pickle.dumps(d)
                tv = self._pickle.loads(hrm)
                if tv.checksum(cypher) != d.checksum(cypher):
                    bads.append((tv, d))

                # test copying
                tv = self._copy.deepcopy(d)
                if tv.checksum(cypher) != d.checksum(cypher):
                    bads.append((tv, d))

            except self._requests.exceptions.ConnectionError as e:
                self._pytest.skip('Internet done goofed')

            # test joblib
            self._lol_joblib(d)

        assert not bads, bads

    def test_malformed(self):
        bads = []
        for i in self.ids_bad:
            try:
                d = self.stream(i)
                d.append(bads)
            except:
                pass

        assert not bads, bads

    def test_hash_eq_id(self):
        hrm = self.ids[0]
        i1 = self.stream(hrm)
        i2 = self.stream(hrm)

        # what did we decide about how to handle this?
        # stream is the same but the state might be different?
        # what are the design tradeoffs here? it means that the
        # stream cannot be used as a dictionary key, which is
        # extremely annoying, but would still have A is B => False

        # we decided (as of now) that all streams hash to their
        # class plus the string representation of their identifier
        # and that equality compares class and identifier equality

        # hash testing
        assert len({i1, i2}) == 1
        assert len({i1.identifier, i2.identifier}) == 1

        assert len({i1, i1.identifier}) == 2
        assert len({i1, i2.identifier}) == 2
        assert len({i1.identifier, i2}) == 2

        # equality testing
        assert i1 == i2
        assert i1.identifier == i2.identifier

        assert i1 != i1.identifier
        assert i1 != i2.identifier
        assert i1.identifier != i2

        # python object identity testing (confusingly)
        assert i1 is not i2
        assert i1.identifier is not i2.identifier

        assert i1 is not i1.identifier
        assert i1 is not i2.identifier
        assert i1.identifier is not i2

    def test_asDict(self):
        for id in self.ids:
            s = self.stream(id)
            try:
                d = s.asDict()
            except self._requests.exceptions.ConnectionError as e:
                self._pytest.skip('Internet done goofed')


class StreamUri(Stream):
    """ Implementation of basic methods for URI streams.

    Probably don't inherit from this class, just reassign them directly
    to any class those primary implementaiton uses uris.
    There are a couple of different ways to do this ...

    Of course then we can't use super() to provide additional information, sigh.
    """

    # TODO consider whether it makes more sense to invert this so that
    # actionable forms simply ask identifiers for their representation
    # in that form ... re: the little dance of asUri -> asUrl
    # asIdentifiersDotOrg, asN2T, asInterpretedByResolverX etc.

    _id_class = None  # NOTE asssigned to idlib.Uri in idlib.__init__


    @property
    def identifier(self):
        return self._identifier

    @property
    def identifier_actionable(self):
        return self.asUri()

    @property  # FIXME pretty sure this shouldn't be properties due to the accept_mimetypes ...
    def id_bound_metadata(self):
        # FIXME this can be id_bound_metadata_free or id_bound_metadata_bound
        # depending on the returned mime type
        # TODO check in on pyontutils.core.OntMetaIri
        raise NotImplementedError
        return self.metadata_free.identifier  # implicitly free?
        return self.data_free.metadata_bound.identifier

    identifier_bound_metadata = id_bound_metadata

    def id_bound_data(self):
        raise NotImplementedError

    _dereference_cache = {}  # FIXME real caching please lru here probably?
    @cache_result  # this is only per instance not globally
    def dereference_chain(self):
        # FIXME this is really dereference chain super
        dc = StreamUri._dereference_cache
        if not self.identifier_actionable in dc:
            try:
                dc[self.identifier_actionable] = tuple(
                    StringProgenitor(resp.url, progenitor=resp)
                    for resp in
                    self._resolution_chain_responses(self.identifier_actionable,
                                                     raise_on_final=False))
            except (exc.ResolutionError, exc.RemoteError) as e:
                # TODO
                # FIXME partial resolution is a complete nightmare
                # to handle >_< ... I don't want to include null
                # pointers here, but we can't just throw an error
                # for a partially resolved chain, I thinkwe have
                # raise with the partial chain as a value in the
                # error
                raise e

        return dc[self.identifier_actionable]

    @cache_result
    def dereference(self, asType=None):
        drc = self.dereference_chain()
        uri = drc[-1]
        if asType:
            nid = asType(uri)
            if isinstance(nid, idlib.Stream):
                if not isinstance(nid._progenitors, dict):
                    # FIXME is are these really progenitors in the way we usually
                    # think of them? ... maybe not?
                    nid._progenitors = {}

                # FIXME TODO do we want/need the full chain here?
                nid._progenitors['id-dereferenced-from'] = self
                # FIXME temporary bridge from StringProgenitor
                nid._progenitors['stream-http'] = uri._progenitor

            return nid
        else:
            # just return the strprg if no type is set
            # don't assume dereferencing is type preserving
            # if we wanted to be really strict about types then we would force
            # explicit conversion to a actionable form and then dereference
            # i.e. ident.actionable().dereference() but that is a pain since
            # the original stream object looses its connection, and the identifier
            # type itself needs to know how to make itself actionable, which it
            # can't quite do in a complete way, in a minimal way probably
            return uri

    def __old_progenitor(self):
        # FIXME there is no single progeneitor here ?
        # or rather, the stream is broken up into objects
        # long before we can do anything about it

        # also metadata headers and data headers are separate
        # possibly even more separate than desired, but of course
        # these aren't single streams, they are bundles of streams
        # so tuple ?
        #self.dereference()
        self.metadata()
        self.data()
        return self.dereference_chain(), self._resp_metadata, self._resp_data

    @cache_result
    def headers(self):
        uri = self.dereference()
        return uri.progenitor().headers

    def data(self, mimetype_accept=None):
        # FIXME TODO should the data associated with the doi
        # be the metadata about the object or the object itself?
        # from a practical standpoint derefercing the doi is
        # required before you can content negotiate on the
        # actual document itself, which is a practical necessity
        # if somewhat confusing
        self._progenitors = {}
        resp = self._requests.get(self.identifier)  # FIXME TODO explicit dereference
        self._progenitors['stream-http'] = resp
        return resp.content

    def asUri(self, asType=None):
        # FIXME this should probably be abstracted to asActionable
        # where actionability is defined by some resolver class
        # defaulting to the system specified default resolver
        # e.g. for doi, orcid, ror, etc. at the moment that is
        # the DNS system for resolving urls
        raise NotImplementedError('implement in subclass')



import os
import unittest
import pytest
import idlib
from idlib.streams import HelpTestStreams

skipif_ci = pytest.mark.skipif('CI' in os.environ, reason='API key required')
SKIP_NETWORK = ('SKIP_NETWORK' in os.environ or
                'FEATURES' in os.environ and 'network-sandbox' in os.environ['FEATURES'])
skipif_no_net = pytest.mark.skipif(SKIP_NETWORK, reason='Skipping due to network requirement')


@pytest.mark.skip('Not ready.')
class TestStreamUri(HelpTestStreams, unittest.TestCase):
    stream = idlib.StreamUri
    ids = [
        'https://github.com',
    ]
    ids_bad = ['lol not an identifier']
    def test_wat(self):
        ic = self.stream._id_class
        i = ic(self.ids[0])


@pytest.mark.skip('Not ready.')
class TestArk(HelpTestStreams, unittest.TestCase):
    stream = idlib.Ark
    ids = [
        'https://doi.org/10.13003/5jchdy',
    ]
    ids_bad = ['lol not an identifier']


class TestDoi(HelpTestStreams, unittest.TestCase):
    stream = idlib.Doi
    ids = [
        'https://doi.org/10.13003/5jchdy',
    ]
    ids_bad = ['lol not an identifier']
    def test_version_f1000(self):
        # this doesn't exist
        #d = idlib.Doi('https://doi.org/10.12688/f1000research.6555')
        #d = d.identifier_bound_version_metadata

        d1 = idlib.Doi('https://doi.org/10.12688/f1000research.6555.1')
        d1.identifier_bound_version_metadata
        # of course there is no linke between the two >_<
        d2 = idlib.Doi('https://doi.org/10.12688/f1000research.6555.2')
        d2.identifier_bound_version_metadata
        # has an 'update-to' field ... which points backwards
        # but how to see if there is something new?


class TestOrcid(HelpTestStreams, unittest.TestCase):
    stream = idlib.Orcid
    ids = [
        'https://orcid.org/0000-0002-1825-0097',
        'https://orcid.org/0000-0001-5109-3700',
        'https://orcid.org/0000-0002-1694-233X',
    ]
    ids_bad = []
    for __i in ids:  # LOL PYTHON class scope list comprehensions
        ids_bad.append(__i.replace('https', 'http'))
    del __i

    def test_validate(self):
        ids = [self.stream._id_class(orcid) for orcid in self.ids]
        bads = [orcid for orcid in ids if not orcid.checksumValid]
        assert not bads, str(bads)

    def test_asType(self):
        import rdflib
        from idlib.formats import rdf as _bind_rdf
        o = idlib.Orcid(self.ids[0])
        nt = o.asType(rdflib.URIRef)
        # still haven't decided if this is a good idea or not
        assert str(o) != str(nt), 'string representation of streams should not match id ???'


class TestRor(HelpTestStreams, unittest.TestCase):
    stream = idlib.Ror
    ids = [
        'https://ror.org/0168r3w48',
        'https://ror.org/02dqehb95',
        'https://ror.org/051fd9666',
    ]
    ids_bad = [
        'https://ror.org/05ht4p406',  # used to exist but api won't tell you that
    ]
    for __i in ids:  # LOL PYTHON class scope list comprehensions
        ids_bad.append(__i.replace('https', 'http'))
    del __i

    def test_init(self):
        ic = idlib.Ror._id_class
        r = idlib.Ror._id_class(prefix='ror.api', suffix='0168r3w48')
        assert type(r.identifier) is str

    def test_validate(self):
        for id in self.ids:
            s = self.stream(id)
            assert s.checksumValid, f'oops, {id}'

    @skipif_no_net
    def test_triples(self):
        from idlib.formats import rdf as _bind_rdf
        r = idlib.Ror(self.ids[0])
        trips = list(r.triples_gen)

    def test_asType(self):
        import rdflib
        from idlib.formats import rdf as _bind_rdf
        r = idlib.Ror(self.ids[0])
        nt = r.asType(rdflib.URIRef)
        assert str(r) != str(nt), 'string representation of streams should not match id'


class TestRrid(HelpTestStreams, unittest.TestCase):
    stream = idlib.Rrid
    ids = [
        'RRID:AB_1234567',
        'RRID:IMSR_JAX:000664',
        'RRID:SCR_001337',
        'RRID:RGD_10395233',
        'RRID:Addgene_19640',
        #'RRID:NCBITaxon_9606',  # TODO not in resolver yet
    ]
    ids_bad = ['lol not an identifier']


@skipif_ci
class TestPio(HelpTestStreams, unittest.TestCase):
    stream = idlib.Pio
    ids = [
        'https://www.protocols.io/view/protocol-for-chronic-implantation-of-patch-electro-b2qgqdtw',
        # api v3 fails for these slugs?
        'https://www.protocols.io/view/protocol-for-chronic-implantation-of-patch-electro-yxmvmno69g3p',
        #'https://www.protocols.io/view/protocol-for-chronic-implantation-of-patch-electro-yxmvmno69g3p/v1',

        'https://www.protocols.io/view/reuse-pc3diyn',
        'https://www.protocols.io/private/8DAE4D2451D5FE18A421D102BC2BEB39',

        #'https://www.protocols.io/view/103',
        #'https://www.protocols.io/view/106',

        'https://www.protocols.io/view/113',
        # changeover happens 128 maybe?
        'https://www.protocols.io/view/136',

        #'https://www.protocols.io/view/622',

        'https://www.protocols.io/view/650',

        'https://www.protocols.io/view/651',
        # changover somewhere in here
        'https://www.protocols.io/view/671',

        'https://www.protocols.io/view/2080',
        # changeover likely at a -> b 2081 -> 2082
        'https://www.protocols.io/view/2088',

        'https://www.protocols.io/view/19358',

        'https://www.protocols.io/view/25122',

        'https://www.protocols.io/view/21417',

    ]
    ids_bad = ['lol not an identifier']

    @skipif_no_net
    def test_users(self):
        for i in self.ids:
            d = self.stream(i)
            if d.creator is not None:
                print(d.creator.orcid)
            for a in d.authors:
                if hasattr(a, 'orcid'):
                    print(a.orcid)

    @skipif_no_net
    def test_id_int(self):
        bads = []
        for i in self.ids:
            un = self.stream(i)
            uh = un.uri_human
            if un.identifier.is_int():
                assert uh.identifier_int == un.identifier.identifier_int

            if uh.identifier_int != uh.identifier.identifier_int:
                bads.append(uh.identifier)

        assert not bads, bads
            
    @skipif_no_net
    def test_org(self):
        #p = idlib.Pio('pio.api:30899')  # not public ?
        #p = idlib.Pio('pio.api:19174')
        #p = idlib.Pio('pio.api:31399')
        p = idlib.Pio('pio.api:25122')  # is public

        #dd = p._data_direct()
        #j = dd.json()['protocol']

        #p1 = idlib.Pio('pio.api1:19174')
        d = p.data()
        d1 = p.data1()
        #breakpoint()
        print(p.asOrg())
        with open('/tmp/ptc.org', 'wt') as f:
            f.write(p.asOrg())


@skipif_ci
class TestPioUser(HelpTestStreams, unittest.TestCase):
    stream = idlib.PioUser
    ids = [
        'pio.user:tom-gillespie',
    ]



import re
from pathlib import Path
import ontquery as oq  # temp implementation detail
import idlib
from idlib import formats
from idlib import streams
from idlib import exceptions as exc
from idlib import conventions as conv
from idlib.cache import cache
from idlib.utils import cache_result, log, base32_crockford_decode
from idlib.config import auth


class _RorPrefixes(conv.QnameAsLocalHelper, oq.OntCuries):
    # set these manually since, sigh, factory patterns
    _dict = {}
    _n_to_p = {}
    _strie = {}
    _trie = {}


_RorPrefixes({'ror': 'https://ror.org/',
             'ror.api': 'https://api.ror.org/organizations/',})


class RorId(oq.OntId, idlib.Identifier):
    _namespaces = _RorPrefixes
    _local_conventions = _namespaces
    # TODO FIXME for ids like this should we render only the suffix
    # since the prefix is redundant with the identifier type?
    # initial answer: yes

    local_regex = r'^0([0-9a-z]{6})([0-9]{2})$'
    canonical_regex = '^https://ror.org/0[0-9a-z]{6}[0-9]{2}$'

    class RorMalformedError(exc.IdlibError):
        """ WHAT HAVE YOU DONE!? """

    def __new__(cls, *args, **kwargs):
        self = super().__new__(cls, *args, **kwargs)
        if self.prefix is None or self.suffix is None:
            raise cls.RorMalformedError(self)
        return self

    @property
    def checksumValid(self):
        """ <https://github.com/ror-community/ror-api/blob/4f91dcb1c4cdeb1c44c92c8c82dc984081585293/
            rorapi/management/commands/convertgrid.py#L11> """

        match = re.match(self.local_regex, self.suffix)
        if match is None:
            return False

        chars, checksum = match.groups()
        number = base32_crockford_decode(chars)
        check = 98 - ((number * 100) % 97)
        return check == int(checksum)

    def _checksum(self, cypher):
        m = cypher()
        m.update(self.identifier.encode())  # down to strings we go!
        return m.digest()


class Ror(formats.Rdf, idlib.HelperNoData, idlib.Stream):

    _id_class = RorId

    identifier_actionable = streams.StreamUri.identifier_actionable
    dereference_chain = streams.StreamUri.dereference_chain
    dereference = streams.StreamUri.dereference
    #progenitor = streams.StreamUri.progenitor
    headers = streams.StreamUri.headers
    #data = idlib.NoDataDereference.data
    #id_bound_data = idlib.NoDataDereference.id_bound_data  # FIXME reuse the Meta and Data from OntRes

    @property
    def checksumValid(self):
        return self._id_class(self.identifier).checksumValid

    @property
    def id_bound_metadata(self):  # FIXME bound_id_metadata bound_id_data
        metadata = self.metadata()
        # wouldn't it be nice if all the metadata schemas had a common field called 'identifier' ?
        id = metadata['id']
        return self._id_class(id)

    identifier_bound_metadata = id_bound_metadata

    @property
    def id_bound_ver_metadata(self):
        return

    identifier_bound_version_metadata = id_bound_ver_metadata

    @cache_result  # FIXME very much must cache these
    def _checksum(self, cypher):  # FIXME unqualified checksum goes to ... metadata ???
        m = cypher()
        metadata = self.metadata()
        name = metadata['name']
        m.update(self.identifier.checksum(cypher))
        m.update(self.id_bound_metadata.checksum(cypher))
        m.update(name.encode())  # unix epoch -> ??
        return m.digest()

    @cache_result
    def metadata(self):
        suffix = self.identifier.suffix
        metadata, path = self._metadata(suffix)
        # oh look an immediate violation of the URI assumption ...
        self._path_metadata = path
        return metadata

    @cache(auth.get_path('cache-path') / 'ror_json', create=True, return_path=True)
    def _metadata(self, suffix):
        # TODO data endpoint prefix ??
        # vs data endpoint pattern ...
        prefix = 'ror.api'  # NOTE THE CHANGE IN PREFIX
        idq = self._id_class(prefix=prefix, suffix=suffix)
        self._resp_metadata = self._requests.get(idq)
        if self._resp_metadata.ok:
            blob = self._resp_metadata.json()
            if len(blob) == 1 and 'errors' in blob:
                errors = blob['errors']
                if len(errors) == 1:
                    error = errors[0]
                    if 'does not exist' in error:
                        # FIXME pretty sure this should be a used to
                        # exist error in the example that causes this
                        raise exc.IdDoesNotExistError(self.identifier)
                    else:
                        raise exc.RemoteError(error)
                else:
                    raise exc.RemoteError(' '.join(errors))
            else:
                return blob
        else:
            try:
                self._resp_metadata.raise_for_status()
            except BaseException as e:
                # FIXME may not be a resolution error
                # if we are raising for status then it is definitely a RemoteError of some kind
                # or the remote telling us that we are in error
                raise exc.RemoteError(self.identifier) from e

    @property
    def name(self):
        return self.metadata()['name']

    def asExternalId(self, id_class):
        eids = self.data['external_ids']
        if id_class._ror_key in eids:
            eid_record = eids[id_class._ror_key]
            if eid_record['preferred']:
                eid = eid_record['preferred']
            else:
                eid_all = eid_record['all']
                if isinstance(eid_all, str):  # https://github.com/ror-community/ror-api/issues/53
                    eid = eid_all
                else:
                    eid = eid_all[0]

            return id_class(eid)

    _type_map = {
        'Education':  'Institution',
        'Healthcare': 'Institution',
        'Facility':   'CoreFacility',
        'Nonprofit':  'Nonprofit',
        'Other':      'Institution',
    }
    @property
    def institutionTypes(self):
        metadata = self.metadata()
        if 'types' in metadata:
            for t in metadata['types']:
                if t == 'Other':
                    log.info(self.label)

                yield self._type_map[t]

        else:
            log.critical(metadata)
            raise TypeError('wat')

    def _triples_gen(self,
                     rdflib=None,
                     rdf=None,
                     rdfs=None,
                     owl=None,
                     NIFRID=None,
                     TEMP=None,
                     **kwargs):
        """ implementation of method to produce a
            triplified version of the record """
        s = self.asType(rdflib.URIRef)
        a = rdf.type
        yield s, a, owl.NamedIndividual  # this goes first in the event the rest fail
        for osuffix in self.institutionTypes:
            o = TEMP[osuffix]
            yield s, a, o

        yield s, rdfs.label, rdflib.Literal(self.label)
        for o in self.synonyms_rdf(rdflib):
            yield s, NIFRID.synonym, o  # FIXME this looses information about synonym type

        # TODO also yeild all the associated grid identifiers

    # normalized fields

    label = name  # map their schema to ours

    def synonyms_rdf(self, rdflib):  # FIXME annoying
        d = self.metadata()
        # FIXME how to deal with type conversion an a saner way ...
        yield from [rdflib.Literal(s) for s in d['aliases']]
        yield from [rdflib.Literal(s) for s in d['acronyms']]
        yield from [rdflib.Literal(l['label'], lang=l['iso639']) for l in d['labels']]

    @property
    def synonyms(self):
        out = []
        m = self.metadata()
        for a in m['aliases'] + m['acronyms']:
            out.append(a)

        for l in m['labels']:
            out.append(l['label'])

        return out

    # alternate representations

    def asUri(self, asType=None):
        return (self.identifier.iri
                if asType is None else
                asType(self.identifier.iri))



import re
import idlib
from idlib import formats
from idlib import streams
from idlib import exceptions as exc
from idlib import conventions as conv
from idlib.cache import cache, COOLDOWN
from idlib.utils import cache_result, log
from idlib.config import auth


class RridId(idlib.Identifier):

    canonical_regex = 'RRID:([A-Za-z]+)_([A-Za-z0-9_:-]+)'
    # NOTE: the canonical regex fails to capture a couple
    # of legacy cases e.g. RRID:WB-STRAIN_ and RRID:MGI:
    # yes I understand why MGI doesn't want RRID:MGI_MGI:
    # but what other identifier namespaces do they use?
    _variant_regex = 'RRID:([A-Z]+):([A-Za-z0-9_:-]+)'  # legacy
    _local_conventions = conv.Identity

    def __init__(self, rrid):
        self._identifier = rrid
        self.authority, self.authority_local_identifier = self.validate()

    def __str__(self):
        return self.identifier

    def validate(self):
        match = re.match(self.canonical_regex, self.identifier)
        if match is None:
            raise exc.MalformedIdentifierError(self.identifier)

        authority, authority_local_identifier = match.groups()
        return authority, authority_local_identifier

    def _checksum(self, cypher):
        m = cypher()
        m.update(self.identifier.encode())
        return m.digest()


class Rrid(formats.Rdf, idlib.HelperNoData, idlib.Stream):

    _id_class = RridId

    _resolver_template = 'https://scicrunch.org/resolver/{id}'

    _COOLDOWN = False

    identifier_actionable = streams.StreamUri.identifier_actionable
    dereference_chain = streams.StreamUri.dereference_chain
    dereference = streams.StreamUri.dereference
    headers = streams.StreamUri.headers

    @property
    def id_bound_metadata(self):  # FIXME bound_id_metadata bound_id_data
        metadata = self.metadata()
        # wouldn't it be nice if all the metadata schemas had a common field called 'identifier' ?
        id = metadata['rrid']['curie']
        return self._id_class(id)

    identifier_bound_metadata = id_bound_metadata

    @property
    def id_bound_ver_metadata(self):
        # RRID records do not have a version at the moment
        # there is a UUID of ambiguous provenace and usefulness
        # but not formal version of the record
        return None

    identifier_bound_version_metadata = id_bound_ver_metadata

    @cache_result
    def metadata(self):
        metadata, path = self._metadata(self.identifier)
        # oh look an immediate violation of the URI assumption ...
        if metadata is not None:
            self._path_metadata = path
            self._progenitor_metadata_blob = metadata
            source = metadata['hits']['hits'][0]['_source']
            return source

    def _cooldown(self):
        self._COOLDOWN = True
        metadata, path = self._metadata(self.identifier)
        return metadata

    @cache(auth.get_path('cache-path') / 'rrid_json', create=True, return_path=True)
    def _metadata(self, identifier):
        idq = self._resolver_template.format(id=identifier)
        #self._resp_metadata = self._requests.get(idq, headers={'Accept': 'application/json'})  # issue submitted
        self._resp_metadata = self._requests.get(idq + '.json')
        if self._resp_metadata.ok:
            return self._resp_metadata.json()
        elif self._COOLDOWN and self._resp_metadata.status_code == 404:
            msg = f'RRID failure: {self._resp_metadata.status_code} {self.asUri()}'
            return {COOLDOWN: msg,}
        else:
            try:
                self._resp_metadata.raise_for_status()
            except BaseException as e:
                raise exc.ResolutionError(identifier) from e

    @cache_result
    def _checksum(self, cypher):
        # FIXME unqualified checksum goes to ... metadata ???
        # TODO figure out what actuall constitues
        # the identity of the RRID record ...
        m = cypher()
        metadata = self.metadata()
        proper_citation = metadata['rrid']['properCitation']
        m.update(self.identifier.checksum(cypher))
        m.update(self.id_bound_metadata.checksum(cypher))
        m.update(proper_citation.encode())
        for vuri in self.vendorUris:
            m.update(vuri.encode())

        return m.digest()

    @property
    def vendorUris(self):
        metadata = self.metadata()
        if 'vendors' in metadata:  # FIXME SCR continues to be a bad citizen >_<
            return [v['uri'] for v in metadata['vendors']]
        else:
            return []

    @property
    def name(self):
        m = self.metadata()
        if m is not None:
            return m['item']['name']

    label = name

    @property
    def synonyms(self):
        m = self.metadata()['item']
        fs = 'label', 'synonyms', 'abbreviations'
        out = []
        for f in fs:
            if f in m:
                for v in m[f]:
                    out.append(v)

        return out

    @property
    def description(self):
        return self.metadata()['item']['description']

    # alternate representations

    def asUri(self, asType=None):
        # TODO n2t, identifiers.org
        # TODO TODO having an explicit model for resolver/metadata services
        # seems like it would subsume the SciGraph/ontquery services
        # along with a bunch of other things ... it would provide
        # proper separation between the implementation details of
        # the identifier classes and their various resolver services
        # this would allow us to sandbox the resolver de jour problem
        uri_string = self._resolver_template.format(id=self.identifier)
        return uri_string if asType is None else asType(uri_string)



import requests
from . import exceptions as exc


def resolution_chain(iri):
    for head in resolution_chain_responses(iri):
        yield head.url


def resolution_chain_responses(iri, raise_on_final=True):
    #doi = doi  # TODO
    s = requests.Session()
    head = requests.head(iri)
    yield head
    while head.is_redirect and head.status_code < 400:  # FIXME redirect loop issue
        yield head.next
        try:
            head = s.send(head.next)
        except requests.exceptions.SSLError as e:
            raise exc.InbetweenError(msg) from e

        yield head
        if not head.is_redirect:
            break

    if raise_on_final:  # we still want the chain ... null pointer error comes later?
        if head.status_code == 404:
            head.raise_for_status()  # probably a permissions issue
        elif head.status_code >= 500:
            raise exc.RemoteError()
        elif head.status_code >= 400:
            msg = f'Nothing found due to {head.status_code} at {head.url}\n'
            raise exc.ResolutionError(msg)



import ontquery as oq  # temp
import idlib
from idlib import families


class IriId(oq.OntId, idlib.Identifier):  # FIXME OntId is only temporary
    """ """
    _family = families.IETF


class UriId(IriId):
    """ """
    _family = families.IETF

    # FIXME code duplication

    # canonicalize the uri


class Uri(idlib.StreamUri):

    _id_class = UriId

    def __gt__(self, other):
        if isinstance(other, idlib.Stream):
            return self.identifier > other.identifier
        else:
            return False  # FIXME TODO

    def asUri(self):
        return self._identifier


class CompactifiedTemplate(Uri):
    _local_conventions = idlib.ConventionsLocal()
    # maybe simpler to implement as a
    # lifting rule over local conventions
    # which raises them to have a global status
    # when parsing strings?
    def __new__(cls, local_or_global):
        # if short make long
        # if long make short etc.
        # probably take from ontquery
        local = local_or_global  # TODO
        global_ = local_or_global  # TODO
        self = super().__new__(cls, global_)
        self.local = local
        # FIXME this is all shorthand that should be rewritten


class Url(Uri):  # FIXME probably just remove this ...
    """ """
    _family = families.IETF





"""
This is ill specified right now. Some thoughts.

The objective is that I have distinguishable content
that is dereferencable from a certain identifier, and
I would like to be able to change some or all of the
content that is returned on dereference.

The underlying semantics for this are extremely diverse.
And there are countless layers of abstraction that you
might have to wade through in order to get to the actual
source of truth.

- Source of truth
  - file on disk mutable
  - file on disk read only but can do an atomic mv/swap
  - file on disk on copy on write file system
  - named path in a git repository at a given ref/rev e.g.
    git-local[conventions]:repo:branch:path/to/file.ext
  - immutable blob in an object store at a given name
  - database (wow this is actually really scary ... it is just
    whatever it is)
- Edit mechanisms
  - Change out the whole source stream behind the scenes
  - Mutate the source stream (what are the atomicity guarantees?)
- Sort of edit
  - Update a known mutable pointer to point to a different immutable pointer.
  - Update a known mutable pointer that is listed in the metadata for
    each immutable pointer.
- Not really edit mechanisms
  - Tell everyone else in the world to change the identifier they use.
  - Add one more pointer to an ever growing deference chain after
    changing the previous referent from a stream to another pointer.
- Stores
  - Mutable store
    - Modify bits in place on a file on a single hard disk.
    - Modify the bits of a "file" in a distributed system,
      if you can do this the abstraction of the distributed
      system is almost surely leaky.
  - Immutable store
    - Update the indirect pointer.
    - Point to a different underlying stream.
    - Change the inode that a particular filename points to.

Furthermore, sometimes you don't actually want to edit the underlying
stream, you just want to update an identifier that is known to point
to a latest version. The issue is that there is a subtle difference in
semantics. Specifically, that in the pure "edit" case there is no
intervening metadata and no recoverable version information, thins
like that. If I have a doi whose type says it dereferences to the
"latest version" of series of dois that dereference immutably.

"""

import idlib


class Stream(idlib.Stream):
    """ example ... """

    def asEditable(self):
        # TODO need a way to set the branch, obtain the repo, find the local file, etc. etc.
        return self._editable_type(self.identifier_editable)

    # in some cases the api for editing/overwriting is simply
    # LocalPath('blah').data = ('lolololol' for _ in range(1))


class EditViaPullRequest:
    pass


class EditViaPush:
    pass


class EditViaApi:
    pass


class EditViaSsh:
    pass


class EditViaRsync:
    pass


class EditViaWebDAV:
    pass


class StreamEdit:
    """ some times you want to edit the contents of a stream
        whose source you may or may not control """

    # XXX just like streams have headers/metadata, we need/want
    # Stream.asEditable or something like that that can capture
    # the process needed to edit/modify the contents of the stream
    # for some types of streams that is always going to be impossible
    # i.e. for immutable streams, others, like files, may simply be a
    # call to open, some might require kicking off a long running process
    # like a pull request and thus need some persistent store to track
    # the the state of the process
    def __init__(self, editable_identifier, static_identifier=None):
        # static identifier would be the source iri
        pass





class IdlibError(Exception):
    """ base class for idlib errors """


class LocalError(IdlibError):
    """ Something went wrong locally and we can do something about it. """


class ConfigurationError(LocalError):
    """ Something went wrong with local configuration. """


class CouldNotReachError(IdlibError):
    """ Could not reach a remote. """
    # FIXME vs UnreachableError, unreachable seems like
    # a more permanent condition

    # consider Aristotle's Comedics
    # permanently unreachable barring another miraculous find of a
    # cache of parchments in a basement
    # that is a no-longer-exists or used-to-exist error
    # where it is in the index but we ...


class CouldNotReachIndexError(CouldNotReachError):
    """ Failure on the way to remote index. """


class CouldNotReachReferentError(CouldNotReachError):
    """ Failure on the way to remote referent.

        This happens after reaching the index and
        the index returns a positive result.

        Note that in a system that is not based on redirects
        this is an error that could be returned by the index
        but it would appear as an referent unknown error. """

    # but its supertype is ReferentUnknown or ReferentUnreachable


class NullReferentError(IdlibError):
    """ You reach some system that gives a difinitive does not
        exist error for a referent that the index says exists.

        This is more or less your friendly null pointer error
        with slightly fewer nasal demons. """


class InbetweenError(IdlibError):
    """ We couldn't get to the remote for some reason. """


class ResolutionError(IdlibError):
    """ something went wrong """
    # FIXME this will be confusing and I'm fairly certain that
    # we should split into NetworkError/TimeoutError
    # and then have RemoteError, because Resolution error is
    # ambiguous, it is the superclass for all of them I think


class MalformedIdentifierError(IdlibError):
    """ Your input cannot be determined to be of the
        type of the identifier you want, therefor we will
        NOT create an object of that type, it cannot exist. """


class IdentifierDoesNotDereferenceToDataError(IdlibError):
    """ Signal that the identifier in question does not
    dereference to data. Examples are physical sample ids,
    or cooluris, ontology class identifiers, etc.

    the ontology class ids are a bit tricky since they do
    dereference, but they do so via a 303, and their dereferenced
    metadata can itself have metadata and data substreams, however
    that should be implemented in the metadata class as we do with
    the ontology resources """
    # 303 see other


class RemoteError(IdlibError):
    """ idlib normalized classes for remote errors """


class ExistenceUnknownError(RemoteError):
    """ For whatever reason the remote cannot provide an
        answer right now. """
    # this abstracts away a whole bunch of implementation details
    # that the user really doesn't need to know about other than
    # maybe check back later


class IdDoesNotExistError(RemoteError):
    """ A record corresponding to the requested identifier does
        not exist on the remote system right now. """
    # 404 not found

class UsedToExistError(RemoteError):
    """ A record corresponding to the requested identifier used
        to exist and will probably never exist again in the future. """
    # 410 gone


class IdExistsButError(RemoteError):
    """ convenience class used to close the space of errors
        where we know the remote exists """


class NotAuthorizedError(IdExistsButError):
    """ the current requesters permissions on the remote system are
        insufficient to grant access to this resource, and what to
        do about it maybe? """
    # this is 403 only NOT 401
    # FIXME 401 =/> that an id exists but 403 does? 401 is you cannot ask that question level/unknown
    # Don't be conused by the fact that HTTP calls these client errors
    # the client didn't do anything wrong


class AccessLimitError(RemoteError):
    """ the remote endpoint is feeling overwhelmed right now or
        you have too many books checked out from the library """
    # 429 too many requests and other such nonsense



"""Identifiers are the smallest possible unit of a stream. Their
fundamental property is that they come equipped with an equality
operation. Not all equality operations are as simple as string= or
numberical equality. These must employ a true identity function that
does not reduce the amount of data that is compared. That is,
identifiers are distinguished from other pieces of data in the sense
that the family of functions that implement 'sameness' requires direct
comparison of every byte of two streams with an allowance for
conversion to a canonical form which may include reordering and
deduplication of elements of the identifier that follow set equality
rather than string equality for example composite primary keys in a
database may be rearranged into a preferred order for further byte to
byte comparison between rows, but the characters in a word cannot be
sorted prior to comparison if we are interested in the equality of two
ordered strings of chars.

Note that under the defintion provided above the ONLY requirement for
an identifier is that it come equipped with an identity function. This
means that whole computer programs can be identifiers as long as the
comparison function is defined. Then there is a question of the robustness
of that identity function to a change in context, specifically defined
as the failures of the equiality of identifiers to correctly imply the
equality of what they dereference to.

There is a tradeoff between robustness of reference and usefulness for
human communication. And for better or for worse the IRBs and IACUCs of
the world tend to frown upon shoving subjects through hash functions.

"""

import idlib


class Identifier(idlib.Stream):  # TODO decide if str should be base ...
    """ Base class for all Identifiers """
    # all identifiers mapped to idlib should have a string representation
    # or be representable as strings in as unambiguous a way as possible
    # this means that distinctions based on types e.g. MyId('1') and YourId('1')
    # need stringify in such a way that they do not colide e.g. myid:1 yourid:1
    # within the expected context of their expansion
    # local:?:myid:1 local:?:yourid:1 could represent the space of local identifiers
    # with unknown local conventions, maintaining a record of all local conventions
    # seems likely to be a giant pain, so local:?: ids would be ephemoral and would
    # probably have to be marked with a source as a kind of best guess maximal domain
    # for assuming similarity, though in some documents every instance of a local:?:id
    # should probably be assumed to be different under expansion

    # as a result of this, it is still not entirely clear whether
    # str is quite the right option, but since normalization can
    # occur before stringification, it is probably ok ...
    # e.g. hello:world -> local:?:hello:world in cases where
    # a local identifier is used without conventions
    # same with local:?:1, local:?:2, local:?:3, local:?:4

    _id_class = str
    local_regex = None  # an unqualified, likely non-unique regex for the system local identifier
    canonical_regex = None  # but really '.+'
    #@staticmethod
    #def normalize(identifier):
        #raise NotImplementedError
        #return identifier

    def __new__(cls, *args, **kwargs):
        return super().__new__(cls, *args, **kwargs)

    def __init__(self, *args, **kwargs):
        return super().__init__(*args, **kwargs)

    #def exists(self):
        # bad identifiers are not allowed to finish __init__
        #raise NotImplementedError

    def metadata(self):
        raise NotImplementedError

    def data(self):
        raise NotImplementedError

    def asLocal(self, conventions=None):
        if conventions is None:
            conventions = self._local_conventions

        return conventions.asLocal(self)





class Rdf:

    def _triples_gen(self,
                     rdflib=None,
                     **kwargs,):
        # add the names you want bound as an explicit kwarg
        raise NotImplementedError

    @classmethod
    def bindRdf(cls, **kwargs):
        cls._triples_gen_kwargs = kwargs

    @property  # FIXME pretty sure this can't be a property if we do this right
    def triples_gen(self):  # FIXME how to do this without rdflib live ... re: pysercomb
        yield from self._triples_gen(**self._triples_gen_kwargs)



from .core import Identity


class QnameAsLocalHelper:
    """ hack for classmethod inheritance sigh """

    @classmethod
    def asLocal(cls, local_or_global):
        return cls.qname(local_or_global)



import logging
from datetime import datetime, timezone
from functools import wraps
from idlib import exceptions as exc


## logging

def makeSimpleLogger(name, level=logging.INFO):
    # TODO use extra ...
    logger = logging.getLogger(name)
    logger.setLevel(level)
    ch = logging.StreamHandler()  # FileHander goes to disk
    fmt = ('[%(asctime)s] - %(levelname)8s - '
           '%(name)14s - '
           '%(filename)16s:%(lineno)-4d - '
           '%(message)s')
    formatter = logging.Formatter(fmt)
    ch.setFormatter(formatter)
    logger.addHandler(ch)
    return logger


log = makeSimpleLogger('idlib')
logd = log.getChild('data')


## time (from pyontutils.utils)

def TZLOCAL():
    return datetime.now(timezone.utc).astimezone().tzinfo


# families (mostly a curiosity)
class families:
    IETF = object()
    ISO = object()
    NAAN = object()


class StringProgenitor(str):
    def __new__(cls, value, *, progenitor=None):
        if progenitor is None:
            raise TypeError('progenitor is a required keyword argument')

        self = super().__new__(cls, value)
        self._progenitor = progenitor
        return self

    def progenitor(self):
        return self._progenitor

    def __getnewargs_ex__(self):
        # have to str(self) to avoid infinite recursion
        return (str(self),), dict(progenitor=self._progenitor)


def cache_result(method):
    """ if the method has run, stash the value for when the method is called again
    WITH NO ARGUMENTS (or at some point the same arguments), last one wins I think
    if you need to clear the cached value CREATE A NEW INSTANCE """

    cache_name = '_cache_' + method.__name__
    @wraps(method)
    def inner(self, *args, **kwargs):
        if not args and not kwargs:
            if hasattr(self, cache_name):
                return getattr(self, cache_name)

        out = method(self, *args, **kwargs)
        setattr(self, cache_name, out)
        return out

    return inner


def makeEnc(alphabet):
    index = {c:i for i, c in enumerate(alphabet)}
    iindex = {v:k for k, v in index.items()}
    base = len(index)

    def encode(number, *, _iindex=iindex, _base=base):
        chars = ''
        while number > 0:
            number, rem = divmod(number, _base)
            chars = _iindex[rem] + chars

        return chars

    def decode(chars, *, _index=index, _base=base):
        number = 0
        for c in chars:
            number = number * _base + _index[c]

        return number


    return encode, decode


# NOTE this assume normalization and downcasing happened in a previous step
# NOTE base32 crockford is big endian under string indexing, * self._base
# acts to move the previous number(s) to the left by one place in that base
crockford_alphabet = '0123456789abcdefghjkmnpqrstvwxyz'
zooko_alphabet     = 'ybndrfg8ejkmcpqxot1uwisza345h769'
pio_alphabet       = 'abcdefghijkmnpqrstuvwxyz23456789'  # start with d is fun
rfc_alphabet       = 'abcdefghijklmnopqrstuvwxyz234567'

(
    base32_crockford_encode,
    base32_crockford_decode,
 ) = makeEnc(crockford_alphabet)

(
    base32_zooko_encode,
    base32_zooko_decode,
) = makeEnc(zooko_alphabet)

(
    base32_pio_encode,
    base32_pio_decode,
) = makeEnc(pio_alphabet)

(
    base32_rfc_encode,
    base32_rfc_decode,
) = makeEnc(rfc_alphabet)



{'config-search-paths': ['{:user-config-path}/idlib/config.yaml',],
 'auth-variables':
 {'cache-path': {'default': '{:user-cache-path}/idlib',
                 'environment-variables': 'IDLIB_CACHE_PATH CACHE_PATH'},
  'log-path': {'default': '{:user-log-path}/idlib',
               'environment-variables': 'IDLIB_LOG_PATH LOG_PATH'},
  'protocols-io-api-creds-file': None,
  'protocols-io-api-store-file': None,
 }}



import ontquery as oq  # temp implementation detail
import idlib
from idlib import formats
from idlib import streams
from idlib import exceptions as exc
from idlib import conventions as conv
from idlib.cache import cache
from idlib.utils import cache_result, log
from idlib.config import auth


class _DoiPrefixes(conv.QnameAsLocalHelper, oq.OntCuries):
    # set these manually since, sigh, factory patterns
    _dict = {}
    _n_to_p = {}
    _strie = {}
    _trie = {}


_DoiPrefixes({'DOI':'https://doi.org/',
              'doi':'https://doi.org/',})


class DoiId(oq.OntId, idlib.Identifier, idlib.Stream):  # also _technically_ a stream
    """ The actual representation of the DOI, but we'll ignore that for now """
    _namespaces = _DoiPrefixes
    _local_conventions = _namespaces
    _id_class = str  # eventually we have to drop down to something
    local_regex = '^10\.[0-9]{4,9}[-._;()/:a-zA-Z0-9]+$'
    canonical_regex = '^https:\/\/doi.org/10\.[0-9]{4,9}[-._;()/:a-zA-Z0-9]+$'

    def __new__(cls, doi_in_various_states_of_mangling=None, iri=None):
        if doi_in_various_states_of_mangling is None and iri is not None:
            doi_in_various_states_of_mangling = iri

        normalize = cls.normalize(doi_in_various_states_of_mangling)
        self = super().__new__(cls, prefix='doi', suffix=normalize)
        self._unnormalized = doi_in_various_states_of_mangling
        self.validate()
        return self

    #@property
    #def identifier(self):
        ## identifier streams return themselves as their identifier
        #return self

    def _checksum(self, cypher):
        m = cypher()
        m.update(self.identifier.encode())
        return m.digest()

    @property
    def metadata(self):
        return None  # FIXME vs identifier?

    @property
    def data(self):
        return None  # FIXME vs identifier?

    @staticmethod
    def normalize(doi):
        doi = doi.replace(' ', '')
        if 'http' in doi or 'doi.org' in doi:
            doi = '10.' + doi.split('.org/10.', 1)[-1]
        elif doi.startswith('doi:'):
            doi = doi.strip('doi:')
        elif doi.startswith('DOI:'):
            doi = doi.strip('DOI:')
        return doi

    @property
    def valid(self):
        return self.suffix is not None and self.suffix.startswith('10.')

    def validate(self):
        if not self.valid:
            raise exc.MalformedIdentifierError(f'{self._unnormalized} does not appear '
                                               f'to be of type {self.__class__}')


# have to exclude, idlib.Uri, idlib.Handle because they are the identifiers NOT the
# streams, and the streams aren't just strings, oops!
# BUT WAIT: maybe we DO want to conflate them!
class Doi(formats.Rdf, idlib.Stream):  # FIXME that 'has canonical representaiton as a uri' issue
    """ The DOI stream. """

    _family = idlib.families.ISO
    _id_class = DoiId

    identifier_actionable = streams.StreamUri.identifier_actionable
    dereference_chain = streams.StreamUri.dereference_chain
    dereference = streams.StreamUri.dereference
    progenitor = streams.StreamUri.progenitor
    headers = streams.StreamUri.headers
    data = streams.StreamUri.data

    def __init__(self, doi_in_various_states_of_mangling=None, iri=None):
        self._identifier = self._id_class(doi_in_various_states_of_mangling, iri)

    def __gt__(self, other):
        if isinstance(other, idlib.Stream):
            return self.identifier > other.identifier
        else:
            return False  # FIXME TODO

    def progenitor(self):
        self.metadata()
        meta = self._resp_metadata if hasattr(self, '_resp_metadata') else self._path_metadata
        return self.dereference_chain(), meta

    @property
    def id_bound_metadata(self):  # FIXME bound_id_metadata bound_id_data
        metadata = self.metadata()
        # wouldn't it be nice if all the metadata schemas had a common field called 'identifier' ?
        URL = metadata['URL']
        DOI = metadata['DOI']
        #prefix = metadata['prefix']  # NOTE NOT the curie meaning of prefix
        return self._id_class(DOI)  # FIXME pretty sure this should just be self.__class__ ?

    identifier_bound_metadata = id_bound_metadata

    @property
    def id_bound_ver_metadata(self):
        # DOIs are the metadat bound version identifier 
        # they run backwards compared to ontology ids
        # by (hopefully) pointing up to a collection
        return None

    identifier_bound_version_metadata = id_bound_ver_metadata

    @property
    def id_bound_data(self):
        data = self.data()  # FIXME mimetype ... from previous? icky
        # beautiful soup this fellow
        return None  # FIXME TODO

    identifier_bound_data = id_bound_data

    @cache_result
    def metadata(self):
        metadata, path = self._metadata(self.identifier)
        # oh look an immediate violation of the URI assumption ...
        self._path_metadata = path
        return metadata

    @cache(auth.get_path('cache-path') / 'doi_json', create=True, return_path=True)
    def _metadata(self, identifier):
        # e.g. crossref, datacite, etc.
        # so this stuff isnt quite to the spec that is doccumented here
        # https://crosscite.org/docs.html
        # nor here
        # https://support.datacite.org/docs/datacite-content-resolver
        accept = (
            'application/vnd.datacite.datacite+json, '  # first so it can fail
            'application/json, '  # undocumented fallthrough for crossref ?
        )
        resp = self._requests.get(identifier, headers={'Accept': accept})
        self._resp_metadata = resp  # FIXME for progenitor
        if resp.ok:
            return resp.json()
        else:
            try:
                self._resp_metadata.raise_for_status()
            except Exception as e:
                if not resp.ok and resp.status_code != 404:  # FIXME control flow here is bad
                    return self._metadata_datacite(identifier)
                else:
                    raise exc.RemoteError(identifier) from e

    def _metadata_datacite(self, identifier):
        """ Try to get data directly from datacite api """
        # FIXME TODO this is why we need a remote/repository class
        # to simplify requesting data on an identifier from a remote
        # what we have here does not compose well at all
        accept = (
            'application/vnd.datacite.datacite+json, '  # first so it can fail
            'application/json, '  # undocumented fallthrough for crossref ?
        )
        datacite_api = f"https://api.datacite.org/dois/{identifier.suffix}"
        resp = self._requests.get(datacite_api, headers={'Accept': accept})
        self._resp_metadata = resp  # FIXME for progenitor
        if resp.ok:
            return resp.json()
        else:
            try:
                self._resp_metadata.raise_for_status()
            except Exception as e:
                raise exc.RemoteError(identifier) from e

    @cache_result  # FIXME very much must cache these
    def _checksum(self, cypher):  # FIXME unqualified checksum goes to ... metadata ???
        m = cypher()
        metadata = self.metadata()
        ts_created = metadata['created']['timestamp']  # key errors inbound I'm sure
        m.update(self.identifier.checksum(cypher))
        m.update(self.id_bound_metadata.checksum(cypher))
        m.update(str(ts_created).encode())  # unix epoch -> ??
        return m.digest()

    # additional streams ...

    def ttl(self):  # this is another potential way to deal with mimetypes
        # both datacite and crossref produce in turtle
        resp = self._requests.get(self.identifier, headers={'Accept': 'text/turtle'})
        self._ttl_resp = resp
        ct = resp.headers['Content-Type']  # FIXME this can KeyError !?
        if 'text/html' in ct:
            # sigh blackfynn
            log.warning(f'{resp.url} is not turtle it is {ct}')  # FIXME duplicate log messages happen here
            return
        else:
            return resp.text

    def metadata_events(self):
        """ metadata about dois from the crossref events api """
        import augpathlib as aug
        events_endpoint = 'https://api.eventdata.crossref.org/v1/events'
        rp = aug.RepoPath(__file__)  # FIXME need __file__ from the calling scope
        try:
            email = rp.repo.config_reader().get_value('user', 'email')
            log.warning(
                (f'your email {email} is being sent to crossref as '
                 'part of the friendly way to use their api'))
            mailto = f'mailto={email}'
        except aug.exceptions.NotInRepoError:
            # TODO failover to the git repo api?
            mailto = 'tgbugs+idlib-no-git@gmail.com'

        handle = self.identifier.suffix
        resp_obj = self._requests.get(f'{events_endpoint}?{mailto}&obj-id={handle}')
        resp_sub = self._requests.get(f'{events_endpoint}?{mailto}&subj-id={handle}')
        # TODO if > 1000 get the rest using the pagination token
        yield from resp_sub.json()['message']['events']
        yield from resp_obj.json()['message']['events']

    # normalized fields

    @property
    def title(self):
        m = self.metadata()
        if 'title' in m:
            return m['title']

        elif 'titles' in m and m['titles']:
            # arbitrary choice to return the first
            return m['titles'][0]['title']

    label = title
    synonyms = tuple()

    @property
    def description(self):
        m = self.metadata()
        breakpoint()

    @property
    def resourceTypeGeneral(self):
        m = self.metadata()
        rtg = 'resourceTypeGeneral' 
        if 'types' in m and rtg in m['types']:
            return m['types'][rtg]

    @property
    def category(self):  # FIXME naming
        """ this is the idlib normalized type of the dereferenced object
        """
            # using category since it matches well with the ontology and registry naming
            # and avoids collisions with type, resourceType, etc.

        rtg = self.resourceTypeGeneral
        if rtg:
            return rtg

        m = self.metadata()
        if 'source' in m and m['source'] == 'Crossref':
            # FIXME sigh ... need representaitons for each
            # type of metadata to avoid this nonsense

            # XXX WARNING the type field on protocols.io records is WRONG
            # dataset was listed because there was no other type that was close
            # so consider that field garbage
            ct = 'container-title'
            if ct in m and m[ct] == 'protocols.io':
                return 'Protocol'

            aj = 'article-journal'
            if 'type' in m and m['type'] == aj:
                return 'ArticleJournal'

    # output streams

    def _triples_gen(self,
                     rdflib=None,
                     rdf=None,
                     rdfs=None,
                     owl=None,
                     NIFRID=None,
                     TEMP=None,
                     **kwargs):
        """ implementation of method to produce a
            triplified version of the record """
        s = self.asType(rdflib.URIRef)
        yield s, rdf.type, owl.NamedIndividual
        try:
            if self.category:
                yield s, rdf.type, rdflib.URIRef(TEMP[self.category])  # FIXME TODO
        except exc.ResolutionError as e:
            log.exception(e)
            yield s, TEMP.resolutionError, rdflib.Literal(True)
            pass

        yield s, rdfs.label, rdflib.Literal(self.label)

    # alternate representations

    def asHandle(self):
        return idlib.Handle(self.identifier.suffix)

    def asUri(self, asType=None):
        return (self.identifier.iri
                if asType is None else
                asType(self.identifier.iri))



import json
import hashlib
import inspect
from pathlib import Path
from functools import wraps
from .utils import log

COOLDOWN = '__idlib.cache_cooldown'

_type_order = (
    bool, int, float, bytes, str, tuple, list, set, dict, object, type, None
)


def type_index(v):
    for i, _type in enumerate(_type_order):
        if isinstance(v, _type):
            return i

    return -1


def args_sort_key(kv):
    """ type ordering """
    k, v = kv
    return k, type_index(v), v


def cache_hash(pairs, cypher=hashlib.blake2s):
    pairs = sorted(pairs, key=args_sort_key)
    converted = []
    for k, v in pairs:
        if k == 'self':  # FIXME convention only ...
            v, _v = v.__class__, v
        converted.append(k.encode() + b'\x01' + str(v).encode())

    message = b'\x02'.join(converted)
    m = cypher()
    m.update(message)
    return m.hexdigest()


def argspector(function):
    argspec = inspect.getfullargspec(function)
    def spector(*args, **kwargs):
        for i, (k, v) in enumerate(zip(argspec.args, args)):
            yield k, v

        if argspec.varargs is not None:
            for v in args[i:]:
                yield argspec.varargs, v

        for k, v in kwargs.items():
            yield k, v

    return spector


def cache(folder, ser='json', clear_cache=False, create=False, return_path=False):
    """ outer decorator to cache output of a function to a folder

        decorated functions accept an additional keyword argument
        `_refresh_cache` that can be used to force refresh the cache """

    if ser == 'json':
        serialize = json.dump
        deserialize = json.load
        mode = 't'
    else:
        raise TypeError('Bad serialization format.')

    write_mode = 'w' + mode
    read_mode = 'r' + mode

    folder = Path(folder)
    if not folder.exists():
        if not create:
            raise FileNotFoundError(f'Cache base folder does not exist! {folder}')

        folder.mkdir(parents=True)

    if clear_cache:
        log.debug(f'clearing cache for {folder}')
        shutil.rmtree(folder)
        folder.mkdir()

    def inner(function):
        spector = argspector(function)
        fn = function.__name__
        @wraps(function)
        def superinner(*args, _refresh_cache=False, **kwargs):
            filename = cache_hash(spector(*args, ____fn=fn, **kwargs))
            filepath = folder / filename
            fe = filepath.exists()
            if fe and not _refresh_cache:
                log.debug(f'deserializing from {filepath}')
                with open(filepath, read_mode) as f:
                    output = deserialize(f)

            else:
                output = function(*args, **kwargs)
                if output is not None:
                    with open(filepath, write_mode) as f:
                        serialize(output, f)

            if isinstance(output, dict) and COOLDOWN in output:
                # a hack to put a dummy variable in the cache
                # to prevent retrying on a persistent failure case
                if fe:
                    reason = output[COOLDOWN]
                    log.debug('currently in cooldown for {args} {kwargs} due to {reason}')

                # TODO add timing or remove cooldowns function
                output = None

            if return_path:
                return output, filepath
            else:
                return output

        return superinner

    return inner



# FIXME this probably makes more sense as an orthauth module?
import sys
import pickle
from urllib import parse
from getpass import getpass
from ..config import auth
from ..utils import log as _log
from . import log as _log

log = _log.getChild('protocols.io')


class ConsoleHelper:

    def run_console_only(self,
                         uri_to_url=True,
                         **kwargs):
        kwargs.setdefault("prompt", "consent")

        try:
            self.redirect_uri = self.client_config['redirect_uris'][0]
        except IndexError as e:
            raise ValueError('No redirect uris were provided in your client config!') from e

        auth_url, _ = self.authorization_url(**kwargs)
        if uri_to_url:
            auth_url = auth_url.replace('redirect_uri', 'redirect_url')

        code = get_auth_code(auth_url)
        self.fetch_token(code=code, include_client_id=True)
        creds = self.credentials
        print('Authentication successful.')
        return creds


def get_auth_code(url):
    import robobrowser
    br = robobrowser.RoboBrowser()
    br.open(url)
    form = br.get_form(id=0)
    if form is None:
        raise ValueError('No form! Do you have the right client id?')
    print('If you registered using google please navigate to\n'
          'the url below and leave email and password blank.')
    print()
    print(url)
    print()
    print(form)
    print()
    print('protocols.io OAuth form')
    e = form['email'].value = input('Email: ')
    p = form['password'].value = getpass()
    if e and p:
        br.submit_form(form)
        params = dict(parse.parse_qsl(parse.urlsplit(br.url).query))

    elif (not e or not p) or 'code' not in params:
        print('If you are logging in via a 3rd party services\n'
              'please paste the redirect url in the prompt')
        manual_url = input('redirect url: ')
        params = dict(parse.parse_qsl(parse.urlsplit(manual_url).query))
        if 'code' not in params:
            print('No auth code provided. Exiting ...')
            sys.exit(10)

    code = params['code']
    return code


def get_protocols_io_auth(creds_file,
                          store_file=auth.get_path('protocols-io-api-store-file'),
                          # yes reading from the store file here at the top level
                          # increases startup time, however if it is burried then
                          # any config error is deferred until later in runtime
                          SCOPES='readwrite'):
    from google_auth_oauthlib.flow import InstalledAppFlow
    from google.auth.transport.requests import Request
    from idlib import exceptions as exc

    InstalledAppFlowConsole = type('InstalledAppFlowConsole',
                                   (ConsoleHelper, InstalledAppFlow),
                                   {})
    if store_file and store_file.exists():
        with open(store_file, 'rb') as f:
            try:
                creds = pickle.load(f)
            except pickle.UnpicklingError as e:
                # FIXME need better way to trace errors in a way
                # that won't leak secrets by default
                log.error(f'problem in file at path for "protocols-io-api-store-file"')
                raise e
    else:
        creds = None

    if not creds or not creds.valid:
        if creds and creds.expired and creds.refresh_token:
            try:
                creds.refresh(Request())
            except Exception as e:
                raise exc.RemoteError('protocols.io refresh error') from e
        elif creds_file is not None:
            flow = InstalledAppFlowConsole.from_client_secrets_file(creds_file.as_posix(), SCOPES)
            creds = flow.run_console_only()
        else:
            msg = 'missing config entry for creds-file'
            raise exc.ConfigurationError(msg)

        with open(store_file, 'wb') as f:
            pickle.dump(creds, f)

    return creds



from .ark import Ark
from .doi import Doi
from .handle import Handle
from .orcid import Orcid
from .pmid import Pmid
from .ror import Ror
from .rrid import Rrid
from .urn import Urn
from .web import Uri



class Conventions:
    """ local or type conventions """


class Identity(Conventions):

    @staticmethod
    def asLocal(identifier):
        if not isinstance(identifier, str):
            return Identity.asLocal(identifier.identifier)
        else:
            return identifier


class ConventionsType(Conventions):
    """ type conventions """


class ConventionsLocal(Conventions):
    """ Base class for all types of local conventions.
        for identifiers local conventions usually convert
        a local identifer into a globally unique identifier
        by providing some additional context.

        Beyond their use for globalizing local identifiers,
        local conventions are information that an be specified
        dynamically, as part of a single stream and that is
        requried in order to correctly interpret information
        that was originally encoded according to those local
        conventions.  """



import re
from .core import ConventionsType


class ConvTypeBytesHeader(ConventionsType):
    """ bytes header type conventions
        this version is intended for headers that start at
        an unknown offset so pattern matching is required """

    def __init__(self, format, start, stop, sentinel):
        self.format = format
        self.start = start
        self.stop = stop
        self.sentinel = sentinel

    def _findElement(self, element, chunk):
        m = re.search(element, chunk)
        if m is not None:
            # FIXME icky api
            return m.start(), m.end()
        else:
            return None, None

    def findStart(self, chunk):
        return self._findElement(self.start, chunk)

    def findStop(self, chunk):
        return self._findElement(self.stop, chunk)

    def findSentinel(self, chunk):
        return self._findElement(self.sentinel, chunk)



from pathlib import Path
import ontquery as oq  # temporary implementation detail
import idlib
from idlib import streams
from idlib import exceptions as exc
from idlib import conventions as conv
from idlib.cache import cache
from idlib.utils import cache_result, log
from idlib.config import auth


# wiki has claims that Orcids are Isnis
# but I need more information ...


class OrcidPrefixes(conv.QnameAsLocalHelper, oq.OntCuries):
    # set these manually since, sigh, factory patterns
    _dict = {}
    _n_to_p = {}
    _strie = {}
    _trie = {}


OrcidPrefixes({'orcid': 'https://orcid.org/',
               'ORCID': 'https://orcid.org/',
               'orcid.pub.3': 'https://pub.orcid.org/v3.0/',})


class OrcidId(oq.OntId, idlib.Identifier):
    _namespaces = OrcidPrefixes
    _local_conventions = _namespaces
    local_regex = ('^0000-000(1-[5-9]|2-[0-9]|3-[0-4])'
                   '[0-9][0-9][0-9]-[0-9][0-9][0-9]([0-9]|X)$')
    canonical_regex = ('^https://orcid.org/0000-000(1-[5-9]|2-[0-9]|3-'
                       '[0-4])[0-9][0-9][0-9]-[0-9][0-9][0-9]([0-9]|X)$')

    class OrcidMalformedError(exc.IdlibError):
        """ WHAT HAVE YOU DONE!? """

    class OrcidLengthError(OrcidMalformedError):
        """ wrong length """

    class OrcidChecksumError(OrcidMalformedError):
        """ failed checksum """

    def __new__(cls, *args, **kwargs):
        """ because __new__ returns the instance we can't split
            and call both paths through __new__ in the MRO, one
            of them always has to be the first and only superclass
            this is non-obvious and annoying """
        self = super().__new__(cls, *args, **kwargs)
        if self.prefix is None or self.suffix is None:
            # XXX NOTE DO NOT, I REPEAT DO NOT try to specify
            # the parser/regex for acceptable identifier forms
            # here, using normalize, or fromUnsafe, or anything else
            # and pass it directly to construction without returning
            # it first, that makes it impossible to hook into the process
            # and return the suggested normalization, and/or we have to
            # make it possible to mark orcid ids as coming from a dirty
            # source which is also bad ... I don't know how to solve this
            # in a sane way, so for now I'm pushing it back to the datasources
            # to fix the malformed input data, this is probably something
            # that idlib could handle, but I need a clearer understanding of
            # the issue before moving forward with implementing normalization
            # NOTE: yes, doi normalization has already been implemented, but
            # that means source data is never corrected if it doesn't match the
            # canonical regex, etc. same issue here with strict parsing vs loose.
            raise cls.OrcidMalformedError(self)

        return self

    @property
    def checksumValid(self):
        """ see
        https://support.orcid.org/hc/en-us/articles/360006897674-Structure-of-the-ORCID-Identifier
        """

        try:
            *digits, check_string = self.suffix.replace('-', '')
            check = 10 if check_string == 'X' else int(check_string)
            total = 0
            for digit_string in digits:
                total = (total + int(digit_string)) * 2

            remainder = total % 11
            result = (12 - remainder) % 11
            return result == check
        except ValueError as e:
            raise self.OrcidChecksumError(self) from e

    def _checksum(self, cypher):
        m = cypher()
        m.update(self.identifier.encode())
        return m.digest()


class Orcid(idlib.HelperNoData, idlib.Stream):

    _id_class = OrcidId

    identifier_actionable = streams.StreamUri.identifier_actionable
    dereference_chain = streams.StreamUri.dereference_chain
    dereference = streams.StreamUri.dereference
    #progenitor = streams.StreamUri.progenitor
    headers = streams.StreamUri.headers

    @cache_result
    def metadata(self):
        suffix = self.identifier.suffix
        metadata, path = self._metadata(suffix)
        # oh look an immediate violation of the URI assumption ...
        self._path_metadata = path
        return metadata

    @cache(auth.get_path('cache-path') / 'orcid_json', create=True, return_path=True)
    def _metadata(self, suffix):
        # TODO data endpoint prefix ??
        # vs data endpoint pattern ...
        prefix = 'orcid.pub.3'  # NOTE THE CHANGE IN PREFIX
        idq = self._id_class(prefix=prefix, suffix=suffix)
        headers = {'Accept': 'application/orcid+json'}
        self._resp_metadata = self._requests.get(idq, headers=headers)
        if self._resp_metadata.ok:
            return self._resp_metadata.json()

    @property
    def id_bound_metadata(self):  # FIXME bound_id_metadata bound_id_data
        metadata = self.metadata()
        # wouldn't it be nice if all the metadata schemas had a common field called 'identifier' ?
        id = metadata['orcid-identifier']['uri']
        return self._id_class(id)

    identifier_bound_metadata = id_bound_metadata

    @property
    def id_bound_ver_metadata(self):
        # TODO
        return

    identifier_bound_version_metadata = id_bound_ver_metadata

    @cache_result  # FIXME very much must cache these
    def _checksum(self, cypher):  # FIXME unqualified checksum goes to ... metadata ???
        # TODO this is a bad checksum
        m = cypher()
        metadata = self.metadata()
        ts_submission = metadata['history']['submission-date']
        m.update(self.identifier.checksum(cypher))
        m.update(self.id_bound_metadata.checksum(cypher))
        m.update(str(ts_submission).encode())  # unix epoch -> ??
        return m.digest()

    # normalized fields

    @property
    def first_name(self):
        m = self.metadata()
        name = m['person']['name']
        if name:  # FIXME cull?
            gn = name['given-names']
            if gn:
                return gn['value']

    @property
    def last_name(self):
        m = self.metadata()
        name = m['person']['name']
        if name:  # FIXME cull?
            fn = name['family-name']
            if fn:
                return fn['value']

    @property
    def label(self):
        return ' '.join([n for n in (self.first_name, self.last_name) if n is not None])

    @property
    def synonyms(self):
        m = self.metadata()
        out = []
        for on in m['person']['other-names']['other-name']:
            out.append(on['content'])

        return out

    def asUri(self, asType=None):
        return (self.identifier.iri
                if asType is None else
                asType(self.identifier.iri))



import re
from setuptools import setup


def find_version(filename):
    _version_re = re.compile(r"__version__ = ['\"](.*)['\"]")
    last = None  # match python semantics
    for line in open(filename):
        version_match = _version_re.match(line)
        if version_match:
            last = version_match.group(1)

    return last


__version__ = find_version('idlib/__init__.py')

with open('README.md', 'rt') as f:
    long_description = f.read()

org_require = ['beautifulsoup4[html5lib]']
rdf_require = ['pyontutils>=0.1.28']
oauth_require = ['google-auth-oauthlib']
tests_require = (['pytest', 'joblib>=1.1.0'] +
                 org_require +
                 rdf_require +
                 oauth_require)
setup(name='idlib',
      version=__version__,
      description='A library for working with identifiers of all kinds.',
      long_description=long_description,
      long_description_content_type='text/markdown',
      url='https://github.com/tgbugs/idlib',
      author='Tom Gillespie',
      author_email='tgbugs@gmail.com',
      license='MIT',
      classifiers=[
          'Development Status :: 4 - Beta',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8',
          'Programming Language :: Python :: 3.9',
          'Programming Language :: Python :: 3.10',
          'Programming Language :: Python :: 3.11',
          'Programming Language :: Python :: Implementation :: CPython',
          'Programming Language :: Python :: Implementation :: PyPy',
          'Operating System :: POSIX :: Linux',
          'Operating System :: MacOS :: MacOS X',
          'Operating System :: Microsoft :: Windows',
      ],
      keywords=('python persistent identifiers'),
      packages=[
          'idlib',
          'idlib.apis',
          'idlib.conventions',
          'idlib.formats',
          'idlib.systems',
      ],
      python_requires='>=3.6',
      tests_require=tests_require,
      install_requires=['orthauth[yaml,sxpr]>=0.0.13', 'requests'],
      extras_require={'test': tests_require,
                      'org': org_require,
                      'rdf': rdf_require,
                      'oauth': oauth_require,
                     },
      scripts=[],
      entry_points={'console_scripts': [ ],},
     )



import idlib
import rdflib
from . import Rdf
from pyontutils.namespaces import rdf, rdfs, owl, NIFRID, TEMP
# FIXME this is clearly backwards, pyontutils should be importing from here

locs = locals()
kwargs = {n:locs[n] for n in
          ['rdflib',
           'rdf',
           'rdfs',
           'owl',
           'NIFRID',
           'TEMP',]}

# bindRdf to any class that needs it
Rdf.bindRdf(**kwargs)
#idlib.Ror.bindRdf(**kwargs)
#idlib.Doi.bindRdf(**kwargs)



import json
from datetime import datetime
import orthauth as oa
import ontquery as oq  # temporary implementation detail
import idlib
from idlib import apis
from idlib import formats
from idlib import streams
from idlib import exceptions as exc
from idlib import conventions as conv
from idlib.cache import cache, COOLDOWN
from idlib.utils import (log,
                         TZLOCAL,
                         cache_result,
                         base32_pio_encode,
                         base32_pio_decode)
from idlib.config import auth


# from neurondm.simple
class OntTerm(oq.OntTerm):
    """ ask things about a term! """
    skip_for_instrumentation = True

    def __new__(cls, *args, **kwargs):
        self = oq.OntId.__new__(cls, *args, **kwargs)
        self._args = args
        self._kwargs = kwargs
        return self

    def fetch(self):
        newself = super().__new__(self.__class__, *self._args, **self._kwargs)
        self.__dict__ = newself.__dict__
        return self


class URIInstrumentation(oq.terms.InstrumentedIdentifier):  # XXX unused

    @property
    def headers(self):
        """ request headers """
        if not hasattr(self, '_headers'):
            resp = self._requests.head(self.iri)  # TODO status handling for all these
            self._headers = resp.headers

        return self._headers

    @property
    def resolution_chain(self):
        # FIXME what should an identifier object represent?
        # the eternal now of the identifier? or the state
        # that it was in when this particular representation
        # was created? This means that really each one of these
        # objects should be timestamped and that equiality of
        # instrumented objects should return false, which seems
        # very bad for usability ...
        if not hasattr(self, '_resolution_chain'):
            # FIXME the chain should at least be formed out of
            # IriHeader objects ...
            self._resolution_chain = [uri for uri in resolution_chain(self)]

        yield from self._resolution_chain

    def resolve(self, target_class=None):
        """ match the terminology used by pathlib """
        # TODO generic probing instrumented identifier matcher
        # by protocol, domain name, headers, etc.
        for uri in self.resolution_chain:
            pass

        if target_class is not None:
            return target_class(uri)

        else:
            return uri  # FIXME TODO identifier it


class _PioPrefixes(conv.QnameAsLocalHelper, oq.OntCuries):
    # set these manually since, sigh, factory patterns
    _dict = {}
    _n_to_p = {}
    _strie = {}
    _trie = {}


_PioPrefixes({'pio.view': 'https://www.protocols.io/view/',  # XXX TODO .html !!! FINALLY
              'pio.edit': 'https://www.protocols.io/edit/',  # sigh
              'pio.run': 'https://www.protocols.io/run/',  # sigh
              'pio.private': 'https://www.protocols.io/private/',
              #'pio.fileman': 'https://www.protocols.io/file-manager/',  # XXX bad semantics
              #'pio.folders.api': 'https://www.protocols.io/api/v3/folders/',  # XXX bad semantics
              'pio.api': 'https://www.protocols.io/api/v3/protocols/',
              'pio.api1': 'https://www.protocols.io/api/v1/protocols/',
})


class PioId(oq.OntId, idlib.Identifier, idlib.Stream):
    _namespaces = _PioPrefixes
    _local_conventions = _namespaces
    canonical_regex = '^https://www.protocols.io/(view|edit|private|api/v3/protocols)/'

    _slug_0_limit = 128  # 113 < ??? < 136
    _slug_0_m = 8
    _slug_0_b = 66010867

    _slug_1_limit = 671  # 651 < ??? < 671
    _slug_1_m = 8
    _slug_1_b = 280510195

    _slug_2_limit = 2082  # 2080 < 2082 < 2088
    _slug_2_m = 1024 ** 2 + 8  # 1048584
    _slug_2_b = 2111848180

    def __new__(cls, curie_or_iri=None, iri=None, prefix=None, suffix=None):
        if curie_or_iri is None and iri:
            curie_or_iri = iri

        if isinstance(curie_or_iri, idlib.Stream):
            normalized = curie_or_iri.identifier.identifier
        elif isinstance(curie_or_iri, idlib.Identifier):
            normalized = curie_or_iri.identifier
        elif curie_or_iri is not None:
            # FIXME trailing nonsense for abstract etc
            normalized = curie_or_iri.replace('://protocols.io', '://www.protocols.io')
        else:
            normalized = None

        try:
            self = super().__new__(cls,
                                   curie_or_iri=normalized,
                                   iri=iri,
                                   prefix=prefix,
                                   suffix=suffix)
        except PioId.Error as e:
            raise exc.MalformedIdentifierError((normalized, iri, prefix, suffix)) from e

        self._unnormalized = curie_or_iri if curie_or_iri else self.iri
        if self.prefix not in self._local_conventions:
            # protocols io fileman folders are not citable artifacts
            raise exc.MalformedIdentifierError(
                f'Not a protocols.io id: {self}')
        return self

    @property
    def uri_human(self):
        return self.__class__(prefix='pio.view', suffix=self.slug)

    @property
    def uri_api(self):
        #if self.prefix == 'pio.fileman':
            #prefix = 'pio.folders.api'  # XXX prefix won't work here, requires a suffix /ids SIGH
        #else:
            #prefix = 'pio.api'
        return self.__class__(prefix='pio.api', suffix=self.slug)

    @property
    def uri_api1(self):
        return self.__class__(prefix='pio.api1', suffix=self.slug)

    def normalize(self):
        return self

    @property
    def slug(self):
        if self.suffix is None:
            breakpoint()

        if self.suffix == 'None':
            log.warning(f'You have a Pio suffix that is the string \'None\'')

        return self.suffix.rsplit('/', 1)[0] if '/' in self.suffix else self.suffix

    @property
    def slug_tail(self):
        if self.is_int():
            id = self.identifier_int
            if id < self._slug_0_limit:
                # the change happens somewhere between 113 and 136
                # superstition tells me that it is probably at 128 or 127
                # depending on the indexing used
                m = self._slug_0_m
                b = self._slug_0_b  # 66010867
                return base32_pio_encode(id * m + b)
            elif id < self._slug_1_limit:
                # the change happens somewhere between 651 and 671
                #m = 8
                #b = 280510195
                m = self._slug_1_m
                b = self._slug_1_b  # 66010867
                return base32_pio_encode(id * m + b)
            if id < self._slug_2_limit:
                # old impl that chops out the middle a
                m = self._slug_2_m
                b = self._slug_2_b - 1
                #full = base32_pio_encode(id * self._slug_2_m + self._slug_2_b - 1)
                full = base32_pio_encode(id * m + b)
                return full[:3] + full[-3:]
            else:
                m = self._slug_2_m
                b = self._slug_2_b
                return base32_pio_encode(id * m + b)
        else:
            return self.slug.rsplit('-', 1)[-1]  # FIXME private issue ...

    @property
    def identifier_int(self):
        if self.is_int():
            return int(self.suffix)

        elif self.is_private():
            # NOTE private ids are NOT random (which is likely bad)
            # and further more show consecutive behavior
            # which means that one is guessable from other (very bad)
            msg = ("Haven't figured out the equation for private -> id")
            raise NotImplementedError(msg)

        st = self.slug_tail
        if len(st) == 6:
            if st.startswith('b88'):  # FIXME magic number
                m = self._slug_0_m
                b = self._slug_0_b
            elif st.startswith('ims'):  # FIXME magic number
                m = self._slug_1_m
                b = self._slug_1_b
            elif st.startswith('mw9'):  # FIXME magic number
                raise NotImplementedError('lacking examples')
                #m = self._slug_1?_m
                #b = self._slug_1?_b
            elif st.startswith('kx3'):  # FIXME magic number
                raise NotImplementedError('lacking examples')
                #m = self._slug_1?_m
                #b = self._slug_1?_b
            elif st.startswith('iwg'):  # FIXME magic number
                raise NotImplementedError('lacking examples')
                #m = self._slug_1?_m
                #b = self._slug_1?_b
            elif st.startswith('kva'):  # FIXME magic number
                # FIXME make params accessible
                m = 8
                b = 355483379
            else:
                # old impl that removes the middle a
                # this affects identifiers < 2082
                st = st[:3] + 'a' + st[3:]
                m = self._slug_2_m
                b = self._slug_2_b - 1
        elif len(st) >= 12:  # new slug format and version nonsense
            raise NotImplementedError(f'not reverse engineered yet {st!r}')
            if st == 'yxmvmno69g3p':  # XXX the doi slug uses the old way, but internally they have some new way
                # 55784 'b2qgqdtw' 'yxmvmno69g3p' ''
                # 55789 'b2qmqdu6' '14egnzrrzg5d' 'v3' # next public
                # 55798 'b2qwqdxe' '8epv51pz5l1b' 'v5' # new slug used in the doi
                'https://www.protocols.io/view/55790'
                'https://www.protocols.io/view/protocol-for-chronic-implantation-of-patch-electro-yxmvmno69g3p/v1'
                breakpoint()
        else:
            m = self._slug_2_m
            b = self._slug_2_b

        # y = mx + b
        d, r = divmod((base32_pio_decode(st) - b), m)
        if r:  # we are in the strange low number regiem ?
            msg = ("Haven't figured out the equation for slugs of the form "
                    f'{st}. d = {d} r = {r}. The id may also be malformed.')
            raise NotImplementedError(msg)
            #breakpoint()

        return d

    @property
    def uri_api_int(self):
        pid = self.__class__(prefix='pio.api', suffix=str(self.identifier_int))
        if not isinstance(pid._progenitors, dict):
            # FIXME is are these really progenitors in the way we usually
            # think of them? ... maybe not?
            pid._progenitors = {}

        pid._progenitors['id-converted-from'] = self
        return pid

    def _checksum(self, cypher):
        m = cypher()
        m.update(self.identifier.encode())
        return m.digest()

    def is_private(self):
        return self.prefix == 'pio.private'

    def is_int(self):
        return self.prefix in ('pio.api', 'pio.view') and self.suffix.isdigit()


def setup(cls, creds_file=None):
    """ because @classmethod only ever works in a single class SIGH """
    if creds_file is None:
        try:
            creds_file = auth.get_path('protocols-io-api-creds-file')
        except KeyError as e:
            raise TypeError('creds_file is a required argument'
                            ' unless you have it in secrets') from e

    try:
        _pio_creds = apis.protocols_io.get_protocols_io_auth(creds_file)
        cls._pio_header = oa.utils.QuietDict(
            {'Authorization': 'Bearer ' + _pio_creds.token})
    except exc.ConfigurationError as e:
        log.warning(e)
        cls._pio_header = None

    if not hasattr(idlib.Stream, '_requests'):
        idlib.Stream.__new__(cls)


class Pio(formats.Rdf, idlib.Stream):
    """ instrumented protocols """

    _id_class = PioId
    # FIXME defining this here breaks the import chain
    # since protocols.py imports from core.py (sigh)
    _wants_instance = '.protocols.ProtocolData'  # this is an awful pattern
    # but what do you want :/

    identifier_actionable = streams.StreamUri.identifier_actionable
    dereference_chain = streams.StreamUri.dereference_chain
    dereference = streams.StreamUri.dereference
    progenitor = streams.StreamUri.progenitor
    headers = streams.StreamUri.headers

    _setup = classmethod(setup)

    #_checked_whether_data_is_not_in_error = False
    #_data_is_in_error = True
    # we MUST assume that data is in error for all instances by
    # default until they prove otherwise HOWEVER the problem is that
    # you now also need another parameter which is whether you have
    # checked to see if it is NOT in error, sigh maybe in error? sigh
    # this becomes hasattr(self, '_data_in_error) and self._data_in_error

    def __new__(cls, *args, **kwargs):
        # sadly it seems that this has to be defined explicitly
        return super().__new__(cls)

    __new__rest = __new__

    def __new__(cls, *args, **kwargs):
        """ self mutating call once setup style """
        cls._setup()
        cls.__new__ = cls.__new__rest
        return cls(*args, **kwargs)

    def __getnewargs_ex__(self):
        # LOL PYTHON
        # Oh you're approaching __new__ ?!
        # apparently using this pattern with __new__
        # breaks the way that loky deserializes things
        return ((self.identifier,), {})

    def __gt__(self, other):
        if isinstance(other, idlib.Stream):
            return self.identifier > other.identifier
        else:
            return False  # FIXME TODO

    @property
    def slug(self):
        return self.identifier.slug

    @property
    def slug_tail(self):
        return self.identifier.slug_tail

    @property
    def doi(self):
        data = self.data()
        if data:
            doi = data['doi']
            if doi:
                return idlib.Doi(doi)

    @property
    @cache_result  # caching this cuts time in half for 2 calls etc. 5s / 10s over 25k calls
    def uri_human(self):  # FIXME HRM ... confusion with pio.private iris
        """ the not-private uri """
        try:
            if len(self.slug_tail) >= 12:
                hrm = self.__class__(self.identifier.uri_human)
                data = hrm.data1(noh=True)
            else:
                data = self.data3()  # XXX data1 cannot bootstrap itself right now
        except exc.RemoteError as e:
            data = None
            try:
                proj = self.progenitor(type='id-converted-from')
                # it should not be the case that we somehow find a
                # private id here because data would have traversed
                # and found it already and gotten the metadata
                # FIXME doi, other int, private should all not be here
                if not proj.identifier.is_int():
                    return proj
                else:
                    raise e
            except KeyError as e2:
                raise e
        if data:
            uri = data['uri']
            if uri:
                return self.fromIdInit(prefix='pio.view', suffix=uri)

    @property
    def uri_human_html(self):
        return self.fromIdInit(prefix='pio.view', suffix=f'{self.identifier_int}.html')

    id_bound_metadata = uri_human  # FIXME vs uri field
    identifier_bound_metadata = id_bound_metadata

    # I think this is the right thing to do in the case where
    # the identifier is the version identifier and versioning
    # is tracked opaquely in the data/metadata i.e. that there
    # is no collection/conceptual identifier
    id_bound_ver_metadata = id_bound_metadata
    identifier_bound_version_metadata = id_bound_ver_metadata

    @property
    def identifier_int(self):
        try:
            return self.data()['id']
        except exc.RemoteError as e:
            try:
                return self.identifier.identifier_int
            except NotImplementedError as e2:
                # internally it is not implemented
                # externally it is a bad id
                # raise the remote error since that is what consumers of this
                # property expect
                try:
                    raise e from exc.MalformedIdentifierError(self.identifier)
                except Exception as e3:
                    raise e3 from e2

    @property
    def uri_api_int(self):
        idint = self.identifier_int

        if not isinstance(idint, int):
            raise TypeError(f'what the {idint}')

        pid = self.fromIdInit(prefix='pio.api', suffix=str(idint))

        if not isinstance(pid._progenitors, dict):
            # FIXME is are these really progenitors in the way we usually
            # think of them? ... maybe not?
            pid._progenitors = {}

        pid._progenitors['id-converted-from'] = self
        return pid

    # XXXXX WARNING IF YOU CHANGE THIS CLEAR THE CACHE OF v1 urls!
    fields = '?fields[]=' + '&fields[]='.join(
        ( # FIXME TODO need to match this list up to other things we need
            'doi',
            'protocol_name',
            'protocol_name_html',
            'creator',
            'authors',
            'description',
            'link',
            'created_on',
            'last_modified',
            'public',
            'doi_status',
            'materials',
            #'materials_text',
            'version',
            'keywords',
            'steps',
            'child_steps',
            )
    )

    def _get_direct(self, apiuri):#, cache=True):
        if not isinstance(self._progenitors, dict):
            # XXX careful about the contents going stale
            self._progenitors = {}

        # XXX this can only get public protocols because
        # we don't have a sane way to log in as a users
        # to get a logged in bearer token
        resp1 = self._requests.get(self.asUri())
        user_jwt = self._get_user_jwt(resp1)
        headers = {'Authorization': f'Bearer {user_jwt}'}
        # XXX the new version slugs have /v1 and /v3 tails >_< AAAAAAAAAAAAAAAAA
        # XXX this substitution may not work as expected
        gau = (apiuri
               #.replace('www', 'go')
               .replace('/v3/', '/v1/'))
        #if cache:
            #oph = self._pio_header
            #try:
                #self._pio_header = headers
                #return self._get_data(gau + self.fields)
            #finally:
                #self._pio_header = oph
        #else:
        resp = self._requests.get(gau + self.fields, headers=headers)
        return resp


    def _data_direct(self, noh=False):
        if noh:
            # XXX noh assumes that the originating form was a uri_human use at own risk
            return self._get_direct(self.identifier.uri_api)
        else:
            return self.uri_human._get_direct(self.identifier.uri_api)

    def data1(self, fail_ok=False, noh=False):
        # FIXME this depends on data3 for uri_human and cannot be used alone at the moment
        # FIXME also only seems to work for public protocols at the moment?

        # FIXME needs robobrowser or similar to function without issues
        # because there is no oauth for it
        # get data from the api/v1 endpoint which is is needed for unmangled steps
        resp = self._data_direct(noh=noh)
        j = resp.json()
        if 'protocol' not in j:  # FIXME SIGH
            if 'status_code' in j and j['status_code'] in (250, 205):
                message = j['text'] + ' ' + self.identifier.asStr()
                raise exc.NotAuthorizedError(message)

            return None

        data = j['protocol']
        return data

    def data3(self, fail_ok=False):
        if not hasattr(self, '_data'):
            self._data_in_error = True
            if not isinstance(self._progenitors, dict):
                # XXX careful about the contents going stale
                self._progenitors = {}

            apiuri = self.identifier.uri_api
            # FIXME XXX the v3 api is completely busted and is missing a ton of
            # information, for example it contains no information about substeps
            # they seemingly appear at random without indication on the ui
            # FIXME the better way is to use
            # /v1/protocols/{private-token} which gets us what we need more easily
            # however this is not trivial to retrive
            # the /v3/protocols/{id} endpoint returns bad/mangled data and we need
            # the /v1/ data to be able to actually render to org
            blob, path = self._get_data(apiuri)
            if 'stream-http' not in self._progenitors:
                self._progenitors['path'] = path

            if blob is None:
                # XXX NOTE this assumes that path.exists() == True
                # XXX whiiich turns out to be a bad assumption ...
                if not path.exists():
                    msg = f'assumption violation cache {path} dne for {self}'
                    log.critical(msg)
                    # TODO raise error or something, or figure out the
                    # conditions under which this assumption goes bad
                    return

                with open(path, 'rt') as f:
                    blob = json.load(f)

                message = blob[COOLDOWN]
                if 'pio_status_code' not in blob:
                    log.critical(blob)
                    path.unlink()
                    raise NotImplementedError('asdf')

                sc = blob['pio_status_code']
                if sc == 212:  # Protocol does not exist
                    if fail_ok: return
                    raise exc.IdDoesNotExistError(message)
                elif sc in (250, 205):  # access requested, not authorized
                    try:
                        # there might be a private id in the progenitor chain
                        nself = self.progenitor(type='id-converted-from')
                        # FIXME TODO this works, but it would be nice if we
                        # could use this to populate the cache for the public
                        # api identifier as well
                        return nself.data(fail_ok=fail_ok)
                    except KeyError as e:
                        pass

                    if fail_ok: return
                    raise exc.NotAuthorizedError(message)
                elif sc in (429,):  # too many requests
                    if fail_ok: return
                    path.unlink()  # hopefully works next time?
                    raise exc.AccessLimitError(message)
                elif sc in (400,):
                    # probably trying to get a pio.view:*.json url
                    # their api is returning nonsense with 400 since
                    # there is nothing wrong with the request ...
                    if fail_ok: return
                    raise exc.NotAuthorizedError(message)
                else:
                    msg = f'unhandled pio status code {sc}\n' + message
                    raise NotImplementedError(msg)
            else:
                if 'status_code' in blob and 'protocol' in blob:
                    self._status_code = blob['status_code']
                    self._data = blob['protocol']
                elif 'id' in blob:  # not via the api
                    self._status_code = 200
                    self._data = blob
                else:
                    log.error(blob)
                    raise exc.RemoteError('no idea what is going on here')

            self._data_in_error = False

            if self._pio_header is None and not self.identifier.is_int():
                # XXX out of band load the uri api int value
                _uai = self.uri_api_int.identifier.uri_api
                self._hack_hash_value = blob
                self._get_data(_uai)

        return self._data

    # data = data3  # XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX

    def data(self, fail_ok=False):
        if not self.identifier.is_private() and len(self.slug_tail) >= 12:
            # FIXME this inverts what should be happening, which is
            # using this information to return the real uri_human and
            # then using that normalized uri instead of the crazy
            # version slug thing, however, we still want provenance
            # for how we got there, even if we toss the crazy slugs
            self._progenitors = {}  # FIXME yes this blasts progens every time
            uh = self.uri_human
            uai = uh.uri_api_int
            d3 = uai.data3(fail_ok=fail_ok)
            self._progenitors['id-converted-to'] = uai
            if 'path' in uai._progenitors:
                self._progenitors['path'] = uai._progenitors['path']
            elif 'stream-http' in uai._progenitors:
                self._progenitors['stream-http'] = uai._progenitors['stream-http']

            return d3
        else:
            return self.data3(fail_ok=fail_ok)

    def asOrg(self):
        from bs4 import BeautifulSoup

        def ct(text):
            return (text
                    .replace('\xa0', ' ')  # non breaking space, our old friend
                    .replace('\u202F', ' ')  # narrow no break space aka '\xe2\x80\xaf'
                    )

        sections = {}
        def hac(soup, cls):
            return soup.has_attr('class') and cls in soup.attrs['class']
        def hasty(soup, sty):
            return soup.has_attr('style') and sty in soup.attrs['style']

        def convert(soup, ind):
            if isinstance(soup, str):
                return soup.replace('\n', f'\n{ind}')

            tag = soup.name
            attrs = soup.attrs

            if tag == '[document]':
                return '\n'.join(convert(c, ind) for c in soup.children)
            if tag == 'html':
                return '\n'.join(convert(c, ind) for c in soup.children)
            if tag == 'head':
                return ''  # html5lib adds this section
                # raise NotImplementedError(f'well this is awkward\n{soup}')
            if tag == 'body':
                return '\n'.join(convert(c, ind) for c in soup.children)

            if tag == 'div' and hac(soup, 'text-blocks'):
                return '\n'.join(convert(c, ind) for c in soup.children)

            if (tag == 'div' and hac(soup, 'text-block')):
                return ''.join(convert(c, ind) for c in soup.children)

            if tag == 'div':
                # XXX watch out here may skip important div info
                return ''.join(convert(c, ind) for c in soup.children)

            if tag == 'img':
                # FIXME TODO need correct typesetting for these
                return f'{ind}[[imgl:{soup.get("src")}]]'
                #return f'\n{ind}#+attr_html: :style float: center; width: 50%;\n{ind}[[imgl:{soup.get("src")}]]'

            if tag == 'br':
                #return ' '
                return '\n'

            if tag == 'a':
                text = soup.text.strip()
                href = soup.get("href")
                if text:
                    return f'[[{href}][{soup.text}]]'
                elif href == '#':
                   return '{{{empty-link}}} '
                else:
                    breakpoint()
                    return f'[[{href}]]'

            if tag == 'ol':
                #bul = f'{ind}0. '
                #sep = f'\n{bul}'
                return ''.join(
                    f'\n{ind}{i + 1}. ' + convert(c, ind)
                    for i, c in enumerate(soup.children))

            if tag == 'ul':
                bul = f'{ind}- '
                sep = f'\n{bul}'
                return '- ' + sep.join(convert(c, ind) for c in soup.children)

            if tag == 'li':
                return ''.join(convert(c, ind) for c in soup.children)

            if tag == 'table':
                return '\n'.join(convert(c, ind) for c in soup.children)
            if tag == 'tbody':
                return '\n'.join(convert(c, ind) for c in soup.children)
            if tag == 'tr':
                return '| ' + ' | '.join(convert(c, ind) for c in soup.children)
            if tag == 'td':
                return ''.join(convert(c, ind) for c in soup.children).replace('\n', ' ')

            if tag == 'sup':
                text = ''.join(convert(c, ind) for c in soup.children)
                return '^{' + text + '}'  # FIXME this needs to be preceeded by something else

            if tag == 'style':
                return ''

            if tag == 'b':
                text = ''.join(convert(c, ind) for c in soup.children)
                ts = text.strip()
                if ts:
                    return '*' + ts + '*'
                else:
                    return ''

            if tag == 'span':
                childs = list(soup.children)
                text = ''.join(convert(c, ind) for c in childs)
                # FIXME combo
                mu = ''
                if hasty(soup, 'font-weight:bold'):
                    mu += '*'
                if hasty(soup, 'font-style:italic'):
                    mu += '/'
                if mu:
                    ts = text.strip()
                    if ts:
                        # FIXME the bane of trailing whitespace
                        return mu + ts + mu[::-1] + ' '
                    else:
                        return ' '

                return text

            if tag in ('o:p', 'pre', 'code'):
                # FIXME ignoring pre and code for now because what the heck are they doing
                # in the free text of protocols ?!
                # I have no idea what the heck this o:p thing is?
                # TODO skip tags group for other apparently empty tags?
                text = ''.join(convert(c, ind) for c in soup.children)
                if tag in ('pre', 'code'):
                    return f'#+html: <{tag}>\n{text}\n#+html: </{tag}>'
                else:
                    return text

            sigh = 'activation', 'simulation'  # FIXME hardcoded workaround
            # protocols.io does not properly escape/render angle brackets in text
            if tag in sigh:
                text = soup.text.replace('\n', f'\n{ind}')
                return ('\n' + ind +
                        '# WARNING THIS TEXT IS MANGLED DUE TO BAD OUTPUT FROM UPSTREAM\n' +
                        f'{ind}#+begin_mangled :tag {tag}\n{ind}{text}\n{ind}#+end_mangled')

            breakpoint()
            raise NotImplementedError((tag, soup))

        def block(type, source, ind):
            s = BeautifulSoup(source['body'], 'html5lib')
            # XXX protocols.io v3 produces malformed html sigh
            _text = convert(s, ind)
            text = _text.replace('\n', f'\n{ind}')  # FIXME see why we still need this? no?
            return f'{ind}#+begin_{type}\n{ind}{text}\n{ind}#+end_{type}'

        stitle = None
        def format_comp(comp, n, ind='  '):
            nonlocal stitle  # sigh
            title = comp['title']
            type_id = comp['type_id']
            source = comp['source']
            order_id = comp['order_id']

            if type_id == 1:  # description
                desc = source['description']
                s = BeautifulSoup(desc, 'html5lib')
                html = ('\n#+begin_export html\n' +
                        desc.replace('<', '\n<') +
                        '\n#+end_export')
                if order_id == 1:
                    tbs = s.find_all('div', 'text-block')
                    converted = [convert(b, ind) for b in tbs]
                    if not converted:
                        log.warning(f'empty?? {comp}')
                    if not sections[stitle]:
                        sections[stitle] = True
                        counter_set = f'[@{n}] '
                    else:
                        counter_set = ''

                    return f'{n}. {counter_set}' + '\n  '.join(converted)  # + html
                else:
                    return convert(s, ind)

                #ul = s.find_all('ul')
                #if ul:
                    #tbs = s.find_all('div', 'text-block')
                    #[b.find_all('li') if  for b in tbs]
                    #breakpoint()
                #return '0. ' + ct(s.text) + html

            elif type_id == 3:  # amount
                inp = source['title']
                unit = source['unit']
                amount = source['amount']
                return ind + f'{amount} {unit} {inp}'  # FIXME embed

            elif type_id == 4:  # duration
                duration = source['duration']
                unit = 'seconds'  # XXX I assume this is in seconds? no unit given
                return ind + f'{duration} {unit}'  # FIXME embed

            elif type_id == 6:  # section
                stitle = source['title']
                if stitle not in sections:
                    sections[stitle] = False
                    return f'** {stitle}'
                else:
                    return ''

            elif type_id == 7:  # external link
                breakpoint()
                raise NotImplementedError(comp)
            elif type_id == 8:  # software package
                name = source['name']
                dev = source['developer']
                dev = dev if dev else ''
                repo = source['repository']
                repo = repo if repo else ''
                return ind + '{{{software(' + ','.join((name, dev, repo)) + ')}}}'
            elif type_id == 9:  # dataset package
                breakpoint()
                raise NotImplementedError(comp)

            elif type_id == 13:  # comment
                #s = BeautifulSoup(source['body'], 'html5lib')
                #text = convert(s.find_all('div', {'class': 'text-blocks'})[0])
                breakpoint()
                raise NotImplementedError(comp)
                text = None # .replace('\n', f'\n{ind}')
                type = 'comment'  # FIXME TODO block type for typesetting
                return f'{ind}#+begin_{type}\n{ind}{text}\n{ind}#+end_{type}'  # FIXME indent

            elif type_id == 15:  # command package
                name = source['name']
                cname = source['command_name']
                cmd = source['command']
                os_name = source['os_name']
                os_version = source['os_version']
                osl = {'MATLAB': 'octave',}
                lang = osl[os_name] if os_name in osl else os_name

                out = (f'{ind}#+caption: {cmd}\n'
                       f'{ind}#+begin_src {lang} :runtime {os_name} :runtime-version {os_version}\n'
                       f'{ind}{name}\n{ind}#+end_src\n')
                _jd = json.dumps(source, indent=2)
                jd = _jd.replace('\n', f'\n{ind}')
                debug = f'\n{ind}#+begin_src json\n{jd}\n{ind}#+end_src\n'
                return out  # + debug

                breakpoint()
                raise NotImplementedError(comp)

            elif type_id == 17:  # expected result
                return block('expected_result', source, ind)

            elif type_id == 18:  # protocol
                # TODO account for these to see if there are
                # ones that are referenced that we don't already have ...
                # TODO on the site these are embedded directly into
                # the protocol, the issue of course is that I haven't
                # implemented intentation level yet, and transclusion
                # probably makes more sense anyway, also it is quite
                # a bit eaiser to understand what is going on when a
                # repeated protocol is not materialized into the page
                # it also reduces duplicate annotation
                _pid = self.identifier.__class__(
                    prefix='pio.api', suffix=str(source['id']))
                return ind + f'[[{_pid.asStr()}][{source["title"]}]]'

            elif type_id == 19:  # safety
                s = BeautifulSoup(source['body'], 'html5lib')
                _text = convert(s, ind)
                text = _text.replace('\n', f'\n{ind}').strip()  # FIXME see why we still need this?
                type = 'safety'  # FIXME TODO block type for typesetting
                return f'{ind}#+begin_{type}\n{ind}{text}\n{ind}#+end_{type}'

            elif type_id == 20:  # reagent
                name = source['name'].strip().replace('\n', ' ').replace(',','\\,')
                url = source['url']
                sku = source['sku']  # sometimes rrid lurks here
                rrid = source['rrid']
                vendor = source['vendor']
                if 'RRID:' in sku:
                    rrid = sku
                if rrid:
                    # may pop vendor info out elsewhere into materials or something?
                    return ind + '{{{' + f'rrid({name},{rrid})' + '}}}'
                    #return f'[[{url}][{name} ({rrid})]]'  # FIXME proper citation etc.

                vname = vendor['name']  # TODO put in materials section I think
                #vurl = vendor['url']

                # FIXME macro need to escape ,
                if sku and url:
                    return ind + '{{{reagent_skurl(' + name + ',' + sku + ',' + url + ')}}}'
                if url:
                    return ind + '{{{reagent_url(' + name + ',' + url + ')}}}'
                if sku:
                    return ind + '{{{reagent_sku(' + name + ',' + sku + ')}}}'

                if vname == 'Contributed by users':
                    return ind + '{{{reagent(' + name + ')}}}'  # TODO markup

                return ind + '{{{reagentv(' + ','.join((name, vname)) + ')}}}'

            elif type_id == 21:  # step cases
                breakpoint()
                raise NotImplementedError(comp)

            elif type_id == 22:  # goto previous step (ouch)
                breakpoint()
                raise NotImplementedError(comp)

            elif type_id == 23:  # file
                url = source['source']
                name = source['original_name']
                return f'{ind}[[{url}][{name}]]'  # TODO render file type somehow?

            elif type_id == 24:  # temperature
                unit = source['unit']
                value = source['temperature']
                if unit == 'Room temperature':
                    return ind + 'room temperature'
                elif unit == 'On ice':
                    # XXX FIXME on ice and ice cold are NOT always the same thing
                    # ice code has a fuzzy quantity associated with it, but
                    # on ice can mean something quite different in some circumstances
                    return ind + 'ice cold'  # FIXME embed
                elif unit == '°C':
                    return ind + f'{value} {unit}'  # FIXME embed

            elif type_id == 25:  # concentration
                _unit = source['unit']
                if _unit == 'Molarity (M)':
                    unit = 'mol/l'
                elif _unit == 'Micromolar (µM)':
                    unit = 'µmol/l'
                elif _unit == 'Nanomolar (nM)':
                    unit = 'nmol/l'
                elif _unit == 'Mass Percent':
                    unit = '% mass'
                elif _unit == 'Volume Percent':
                    unit = '% volume'
                elif _unit == 'Mass/Volume Percent':
                    # what in the name of all unholy dimensionalities is this !?!?
                    unit = '% mass/volume'
                else:
                    log.warning(f'unhandled unit {_unit}')
                    unit = _unit

                value = source['concentration']
                return ind + f'{value} {unit}'  # FIXME embedding :/

            elif type_id == 26:  # notes
                s = BeautifulSoup(source['body'], 'html5lib')
                _text = convert(s.find_all('div', {'class': 'text-blocks'})[0], ind)
                text = _text.replace('\n', f'\n{ind}')
                type = 'note'  # FIXME TODO block type for typesetting
                return f'{ind}#+begin_{type}\n{ind}{text}\n{ind}#+end_{type}'

            elif type_id == 28:  # equipment
                brand = source['brand']
                _sku = source['sku']
                sku = '' if _sku == 'N/A' or not _sku else _sku + ' '
                name = source['name']
                # TODO markup somehow maybe ?
                return ind + f'{name} ({sku}{brand})'

                breakpoint()
                raise NotImplementedError(comp)

            elif type_id == 30:  # centrifuge
                # TODO apparently there is a units field that can
                # be pulled back from a public protocol that is needed
                # to translate this section (WAT)
                if not hasattr(self.__class__, '_pio_units'):
                    # FIXME a bit more systematic approach perhaps
                    resp = self._requests.get(
                        'https://www.protocols.io/api/v3/units',
                        headers=self._pio_header)
                    j = resp.json()
                    self.__class__._pio_units = j['units']

                value = source['centrifuge']
                unit_id = source['unit']
                unit = 'times-earth-gravity' if unit_id == 34 else '???'

                # why the heck did they create yet another thing that
                # has an embedded temperature !?
                temp_unit_id = source['temperatureUnit']
                temp_unit = 'K' if temp_unit_id == 10 else '???'
                # there is no good way to do this ... 1000 m/s2
                return ind + f'{value} {unit}'

            breakpoint()
            raise NotImplementedError(comp)

        def format_step(step, n):
            comps = []
            for c in sorted(step['components'], # reverse=True,
                            key=lambda c: c['order_id']):
                fc = format_comp(c, n)
                comps.append(fc)

            # FIXME big problems here
            return '\n'.join(comps)

        def format_mats(mats):
            # TODO molecular weight?
            hs = 'name', 'sku', 'vendor name', 'chemical name', 'cas number', 'rrid', 'url'
            table = '\n| ' + ' | '.join(hs) + '\n|-'
            for m in mats:
                vs = (m['name'], m['sku'], m['vendor']['name'],
                      m['linfor'],  # FIXME is this really chemical name?
                      m['cas_number'], m['rrid'], m['url'])
                row = [v.strip() if isinstance(v, str) else '' for v in vs]
                table += '\n| ' + ' | '.join(row)

            return table

            #return ('\n#+begin_src json\n' +
                    #json.dumps(mats, indent=2) +
                    #'\n#+end_src')
            #for mat in mats:

        import htmlfn as hfn
        data = self.data()
        # FIXME may encounter anchoring issuse if we do this
        # will likely need to adjust accordingly
        lrc = f'<link rel="canonical" href="{self.uri_api_int.asStr()}">'
        doi = self.doi
        if doi:
            metas = (
                dict(name='dc.identifier', content=doi.asStr()),
                dict(name='DOI', content=doi.asStr()),
                    )
        else:
            metas = tuple()

        metas += (
            dict(name='og.url', content=self.uri_human.asStr()),
        )

        helper = [s for s in data['steps'] # debug
                  if [c for c in s['components'] if c['type_id'] in (24, 4, 3)]]

        _war = data['warning']
        _waf = convert(BeautifulSoup(_war, 'html5lib'), '').strip() if _war else ''
        warnings = ('\n* Warnings\n' + _waf if _waf else '')

        _gls = data['guidelines']
        _glf = convert(BeautifulSoup(_gls, 'html5lib'), '').strip() if _gls else ''
        guidelines = ('\n* Guidelines\n' + _glf if _glf else '')

        _mat = data['materials']
        # TODO add additional materials accumulated from inline content
        # no idea why pio doesn't do that automatically itself ...
        materials = ('\n* Materials\n' + format_mats(_mat).strip()
                     if _mat else '')

        body = (
            f'\n\n[[{self.uri_human.asStr()}][Original on protocols.io]]\n' +
            # guidelines warnings materials metadata
            warnings +
            guidelines +
            materials +
            '\n* Steps\n' +
            '\n'.join(format_step(s, i + 1) for i, s in enumerate(data['steps'])))

        title = data['title']

        options = (
            '\n#+options: ^:nil'
            '\n'
        )

        # TODO probably want a span macro
        macros = (
            '\n#+macro: empty-link *ERROR-UPSTREAM* '
            '\n#+macro: rrid =$1 ($2)='
            '\n#+macro: reagent_skurl =$1 ($2)= @@comment: $3@@'
            '\n#+macro: reagent_sku =$1 ($2)='
            '\n#+macro: reagent_url =$1= @@comment: $2@@'
            '\n#+macro: reagentv =$1 ($2)='
            '\n#+macro: reagent =$1='
            '\n#+macro: software [[$3][$1]] @@comment: $2@@'
            '\n'
        )

        doc = (
            f'#+title: {title}\n' +
            options +
            macros +
            '#+html_head: <style>img { float: center; width: 50%; }</style>\n'
            f'#+html_head: {lrc}\n' +
            '\n'.join(['#+html_head: ' + hfn.metatag(**md) for md in metas]) +
            body)

        return ct(doc)

        #return hfn.htmldoc(body,
                           #title=data['title'],
                           #other=(lrc,),
                           #metas=metas,
                           #styles=tuple(),)

    @staticmethod
    def _get_user_jwt(resp):
        """ an aweful way to get this that surely will break """
        text = resp.text
        before, after = text.split('USER_JWT')
        eq, user_jwt, rest = after.split('"', 2)
        return user_jwt

    @cache(auth.get_path('cache-path') / 'protocol_json', create=True, return_path=True)
    def _get_data(self, apiuri):
        """ use apiuri as the identifier since it is distinct
            from other views of the protocol e.g. uri_human etc. """

        if hasattr(self, '_hack_hash_value') and self._hack_hash_value is not None:
            # make it possible to cache an arbitrary value without
            # actually retrieving it
            v = self._hack_hash_value
            self._hack_hash_value = None
            return v

        # TODO progenitors
        log.debug('going to network for protocols')
        if self._pio_header is None and isinstance(self, Pio):
            # the hacks in this branch only apply to the Pio class
            # otherwise always use apiuri in the other branch
            if self.identifier.is_private():
                resp = self._get_direct(apiuri)#, cache=False)
            else:
                if self.identifier == self.identifier.uri_api_int:
                    try:
                        prog = self.progenitor(type='id-converted-from')
                        # FIXME pretty sure this should return None instead of error?
                    except KeyError as e:
                        msg = 'no protocols io api credentials'
                        raise exc.NotAuthorizedError(msg) from e

                    slug = prog.slug
                else:
                    slug = self.slug

                hack = self._id_class(
                    prefix='pio.view',
                    suffix=slug).asStr() + '.json'
                resp = self._requests.get(hack)
        else:
            resp = self._requests.get(apiuri, headers=self._pio_header)
        #log.info(str(resp.request.headers))
        self._progenitors['stream-http'] = resp
        if resp.ok:
            try:
                j = resp.json()  # the api is reasonably consistent
                return j
            except Exception as e:
                log.exception(e)
                raise e
        else:
            try:
                j = resp.json()
                sc = j['status_code']
                em = j['error_message']
            except Exception as e:
                sc = resp.status_code
                em = resp.reason

            msg = (f'protocol issue {self.identifier} {resp.url} '
                   f'{resp.status_code} {sc} {em}')
            self._failure_message = msg  # FIXME HACK use progenitor instead
            return {COOLDOWN: msg,
                    'http_status_code': resp.status_code,
                    'pio_status_code': sc,
                    'error_message': em,}

    metadata = data  # FIXME

    @cache_result
    def _checksum(self, cypher):
        m = cypher()
        # FIXME TODO hasing of python objects ...
        metadata = self.metadata()
        #m.update(self.identifier.checksum(cypher))
        # XXX self.identifer cannot be included because
        # it makes it impossible to dealias tha various different referents
        m.update(self.id_bound_metadata.identifier.checksum(cypher))
        #m.update(self.version_id)  # unix epoch -> ??
        m.update(self.updated.isoformat().encode())  # in principle more readable
        #m.update(self.updated.timestamp().hex())
        return m.digest()

    @property
    def hasVersions(self):
        return bool(self.data()['has_versions'])

    @property
    def versions(self):
        yield from self.data()['versions']  # TODO ...

    @property
    def created(self):
        # FIXME I don't think this is TZLOCAL for any reason beyond accident of circumstances
        # I think this is PDT i.e. the location of the protocols.io servers
        tzl = TZLOCAL()
        return datetime.fromtimestamp(self.data()['created_on'], tz=tzl)

    @property
    def updated(self):
        tzl = TZLOCAL()
        return datetime.fromtimestamp(self.data()['changed_on'], tz=tzl)

    @property
    def title(self):
        data = self.data()
        if data:
            title = data['title']
            if title:
                return title

    label = title

    @property
    def label_safe(self):
        """don't fail if data access is missing """
        try:
            return self.label
        except exc.RemoteError:
            return self.identifier.slug

    @property
    def creator(self):
        username = self.data()['creator']['username']
        if username is not None:
            return PioUser('pio.user:' + username)

    @property
    def authors(self):
        class Author:
            def __init__(self, blob):
                self.blob = blob
                self.name = blob['name']

        for u in self.data()['authors']:
            yield Author(u)
            continue
            # FIXME TODO
            _username = u['username']
            username = (_username
                        if _username is not None else
                        (u['name'].replace(' ', '-') + 'FAKE'))
            uid = PioUserId(prefix='pio.user', suffix=username)
            pu = PioUser(uid)
            if _username is None:
                def metadata(self, __asdf=u):
                    return __asdf

            yield pu

    def asUri(self, asType=None):
        return (self.identifier.iri
                if asType is None else
                asType(self.identifier.iri))

    def asDict(self, include_description=False, include_private=True):
        """ XXX this should NEVER allow an error to escape.
            Only return less information. """
        if self.identifier.is_int():
            out = super().asDict(include_description)

            try:
                out['uri_human'] = self.uri_human.identifier  # prevent double embedding
            except exc.RemoteError as e:
                pass

            if hasattr(self, '_data_in_error') and self._data_in_error:
                return out

            # NOTE if you started from a doi then it seems extremely unlikely
            # that you would be in a sitution where data retrieval could fail
            # which means that really only the uri_human case can fail and
            # there still be a chance that there is a uri_human we can use
            doi = self.doi
            if doi is not None:
                out['doi'] = doi
            return out
        else:
            try:
                uri_api_int = self.uri_api_int
                if uri_api_int is None:
                    # This should trigger a remote error, if not, we want to
                    # know because something very strange is going on
                    self.data()

                out = uri_api_int.asDict(include_description)
                if include_private and self.identifier.is_private():
                    out['uri_private'] = self.identifier  # FIXME some way to avoid leaking these if needed?
                return out
            except exc.RemoteError as e:
                # we don't have any metadata but we will return what little info we have
                return super().asDict(include_description)

    def _triples_gen(self,
                     rdflib=None,
                     rdf=None,
                     rdfs=None,
                     owl=None,
                     NIFRID=None,
                     TEMP=None,
                     **kwargs):

        s = self.asType(rdflib.URIRef)

        yield s, rdf.type, owl.NamedIndividual

        if self.uri_human:
            # XXX dereference checks should not be run here, they
            # should be conduceded centrally during
            yield s, TEMP.hasUriHuman, self.uri_human.asType(rdflib.URIRef)

        if self.label:
            yield s, rdfs.label, rdflib.Literal(self.label)

        doi = self.doi
        if doi is not None:
            yield s, TEMP.hasDoi, doi.asType(rdflib.URIRef)


class _PioUserPrefixes(conv.QnameAsLocalHelper, oq.OntCuries):
    # set these manually since, sigh, factory patterns
    _dict = {}
    _n_to_p = {}
    _strie = {}
    _trie = {}


_PioUserPrefixes({'pio.user': 'https://www.protocols.io/researchers/',
                  'pio.api.user': 'https://www.protocols.io/api/v3/researchers/',
                  'pio.api1.user': 'https://www.protocols.io/api/v1/researchers/',
})


class PioUserId(oq.OntId, idlib.Identifier, idlib.Stream):
    _namespaces = _PioUserPrefixes
    _local_conventions = _namespaces

    _checksum = PioId._checksum

    @property
    def uri_human(self):
        return self.__class__(prefix='pio.user', suffix=self.suffix)

    @property
    def uri_api(self):
        return self.__class__(prefix='pio.api.user', suffix=self.suffix)

    @property
    def uri_api1(self):
        return self.__class__(prefix='pio.api1.user', suffix=self.suffix)


class PioUser(idlib.HelperNoData, idlib.Stream):
    _id_class = PioUserId

    _get_data = Pio._get_data
    asUri = Pio.asUri
    _setup = classmethod(setup)

    identifier_actionable = streams.StreamUri.identifier_actionable
    dereference_chain = streams.StreamUri.dereference_chain
    dereference = streams.StreamUri.dereference
    progenitor = streams.StreamUri.progenitor
    headers = streams.StreamUri.headers

    def __new__(cls, *args, **kwargs):
        # sadly it seems that this has to be defined explicitly
        return object.__new__(cls)

    __new__rest = __new__

    def __new__(cls, *args, **kwargs):
        """ self mutating call once setup style """
        cls._setup()
        cls.__new__ = cls.__new__rest
        return cls(*args, **kwargs)

    def __getnewargs_ex__(self):
        # LOL PYTHON
        # Oh you're approaching __new__ ?!
        # apparently using this pattern with __new__
        # breaks the way that loky deserializes things
        return ((self.identifier,), {})

    @property
    def uri_human(self):
        return self.__class__(self.identifier.uri_human)

    @property
    def uri_api(self):
        return self.__class__(self.identifier.uri_api)

    @property
    def uri_api1(self):
        return self.__class__(self.identifier.uri_api1)

    id_bound_metadata = uri_human
    identifier_bound_metadata = id_bound_metadata
    id_bound_ver_metadata = id_bound_metadata
    identifier_bound_version_metadata = id_bound_ver_metadata

    @property
    def _id_act_metadata(self):
        """ the actionalble identifier that is directly used to
            retrieve metadata from the resolver

            TODO there is a better way to implement this which is
            split the metadata streams into their own classes
            where the metadata is now treated as data """
        # FIXME this isn't actuall the id_act
        # it is really just a convenient way to
        # skip dealing with the send_* arguments for
        # metadata and data, namely that the query argument
        # that we append here should be explicit either
        # as a default kwarg or something similar
        # the fact that query parameters can sometimes
        # be part of the identity of a thing vs used to
        # modify exactly which parts of a thing are returned
        # is problematic, but is also something that needs
        # to be tracked
        return self.uri_api1.asUri() + '?with_extras=1'

    @cache_result
    def metadata(self):
        self._progenitors = {}
        blob, path = self._get_data(self._id_act_metadata)
        if 'stream-http' not in self._progenitors:
            self._progenitors['path'] = path

        if blob is not None:
            self._status_code = blob['status_code']
            return blob['researcher']
        else:
            return {'name': ''}

    def _metadata_refresh(self):
        # FIXME nasty code reuse issues here especially
        # around the fact that _refresh_cache can't be
        # passed along the scope unless the decorated function
        # does it manually, which kinda defeats the point

        _old_prog = self._progenitors
        self._progenitors = {}
        for k in ('id-converted-from',): # TODO generalize to persistent progs
            if k in _old_prog:
                self._progenitors[k] = _old_prog[k]

        blob, path = self._get_data(self._id_act_metadata, _refresh_cache=True)
        return self.metadata()

    @cache_result  # FIXME XXX cache_result currently does not work if there are arguments!
    def _checksum(self, cypher):
        m = cypher()
        metadata = self.metadata()
        m.update(self.identifier.checksum(cypher))
        m.update(self.id_bound_metadata.identifier.checksum(cypher))
        orcid = self.orcid
        if orcid is not None:
            # use orcid.identifier.checksum to prevent
            # changes to the orcid record from causing
            # changes in the identity of the user record
            m.update(orcid.identifier.checksum(cypher))

    @property
    def orcid(self):
        orcid = self.metadata()['orcid']
        if orcid is not None:
            return idlib.Orcid.fromIdInit(prefix='orcid', suffix=orcid)

    @property
    def name(self):
        return self.metadata()['name']

    label = name

    @property
    def bio(self):
        return self.metadata()['bio']

    description = bio


class IsniId(oq.OntId):  # TODO
    prefix = 'http://isni.org/isni/'
    _ror_key = 'ISNI'


class Auto:
    """ dispatch to type on identifier structure """
    def __new__(cls, something):
        if '10.' in something:
            if 'http' in something and 'doi.org' not in something:
                pass  # probably a publisher uri that uses the handle
            else:
                return idlib.Doi(something)

        if 'orcid' in something:
            return idlib.Orcid(something)

        if '/ror.org/' in something or something.startswith('ror:'):
            return idlib.Ror(something)

        if 'protocols.io' in something:
            return idlib.Pio(something)

        try:
            return oq.OntId(something)
            return OntTerm(something)  # use the better local version of OntTerm
        except oq.OntId.Error as e:
            raise exc.MalformedIdentifierError(something) from e




from idlib import exceptions as exc
from idlib.utils import families
from idlib.streams import (Stream,
                           HelperNoData,
                           StreamUri,)
from idlib.identifiers import (Identifier,)

import idlib.conventions.type  # LOL PYTHON
import idlib.conventions.local  # LOL PYTHON
from idlib.conventions.core import ConventionsLocal
from idlib.conventions.local import (Curies,
                                     QNames,)
from idlib.from_oq import (Auto,
                           Pio,
                           PioUser,)
from idlib.systems import (Ark,
                           Doi,
                           Handle,
                           Orcid,
                           Pmid,
                           Ror,
                           Rrid,
                           Uri,
                           Urn,)

# assign default identifier classes to streams
StreamUri._id_class = Uri

def get_right_id(uri):
    # FIXME this is a bad way to do this ...
    class StreamUriTemp(StreamUri):
        def asUri(self):
            return self.identifier

    if isinstance(uri, Doi) or 'doi' in uri and '/doi/' not in uri:
        if isinstance(uri, Doi):
            di = uri
        elif 'doi' in uri:
            di = Doi(uri)

        try:
            pi = di.dereference(Pio)
        except exc.MalformedIdentifierError:
            pi = di.dereference(StreamUriTemp)

    else:
        if not isinstance(uri, Pio):
            try:
                pi = Pio(uri)  #.normalize()
            except exc.MalformedIdentifierError:
                pi = StreamUriTemp(uri)
        else:
            pi = uri

    return pi


__version__ = '0.0.1.dev14'
